from math import pi, sqrt

from ase.data import atomic_numbers, reference_states, chemical_symbols

from asap3 import EMT, EMT2013
from asap3.Internal.EMTParameters import _std_emt_params

beta = (16.0 * pi / 3.0)**(1.0 / 3.0) / sqrt(2.0)

def EMTFit(elements, parameters):
    """Create an EMT object from a list (not dict!) of parameters."""
    #print "EMTFit called with:"
    #print "  elements =", elements
    #print "  parameters =", parameters
    parameternames = ['eta2', 'lambda', 'kappa', 'E0', 'V0', 'S0', 'n0']
    param = {}
    for k, v in parameters.items():
        if isinstance(k, tuple):
            k, k2 = k
            assert k == k2
        p = {}
        for name, value in zip(parameternames, v):
            p[name] = value
        # Check for resonable parameters
        #if p['eta2'] * 1.809 - p['kappa'] < 0.01:
        #    return None
        param[k] = p
    #print "  new parameters:", param
    return EMT(param)

def EMT2013Fit(elements, parameters, order='kappa'):
    """Create an EMT2013 object from a list (not dict!) of parameters."""
    #print "EMT2013Fit called with:"
    #print "  elements =", elements
    #print "  parameters =", parameters
    if order == 'kappa':
        parameternames = ['eta2', 'lambda', 'kappa', 'E0', 'V0', 'S0', 'n0']
    elif order == 'delta':
        parameternames = ['eta2', 'lambda', 'delta', 'E0', 'V0', 'S0', 'n0']
    else:
        raise ValueError('"order" must be either "kappa" or "delta"')
    limits = (('eta2', 0.1, 25.0),
              ('lambda', 0.1, 25.0),
              ('kappa', 0.2, 50.0),
              ('E0', -25, 0.0),
              ('V0', 0.0, 12.0),
              ('S0', 0.5, 5.0),
              ('n0', 0.0, 1.0))
    
    param = {'Calculator': "EMT2013_V1"}
    for k, v in parameters.items():
        if isinstance(k, tuple):
            k, k2 = k
            assert k == k2

        # Copy parameters
        p = {}
        if isinstance(v, dict):
            p = v.copy()
        else:
            p = dict(zip(parameternames, v))

        # Convert 'delta' to 'kappa'
        if order == 'delta':
            p['kappa'] = beta * p['eta2'] - p['delta']
            del p['delta']

        # Check for resonable parameters
        if beta * p['eta2'] - p['kappa'] < 0.01:
            raise ValueError(str(k) + ':  beta * eta2 - kappa < 0.01')
        for name, lower, higher in limits:
            if p[name] < lower or p[name] > higher:
                raise ValueError('%s(%s) = %.5f is not in [%.5f, %.5f]' 
                                 % (name, str(k), p[name], lower, higher))
        param[k] = p
    #print "  new parameters:", param
    return EMT2013(param)

EMT2011Fit = EMT2013Fit

def EMTStdParameters(z, order='kappa', guess=False, ecoh={}):
    """Get the standard EMT parameters, to be used as a starting guess for optimization.
    
    Parameters:
    z:  Atomic number or chemical symbol.
    order: Whether kappa or delta is on the list.  Default: kappa.
    guess: If true, values are guesstimated for unknown elements.
    ec: If guessing then E0 anc V0 can be based on cohesive energies from this dict.
        Keys are chemical symbols, values are cohesive energies in eV.
    
    Returns:
    A list of parameters in the order [eta2, lambda, kappa or delta, E0, V0, S0, n0].
    """
    if isinstance(z, str):
        z = atomic_numbers[z]

    parameternames = ['eta2', 'lambda', 'kappa', 'E0', 'V0', 'S0', 'n0']
    if z in _std_emt_params and guess != 'force':
        # Known in std EMT
        parameters = [_std_emt_params[z][x] for x in parameternames]
    elif guess:
        # eta2, lambda and kappa vary little amongst the elements, an not in a systematic way.
        # Use average values
        parameters = [3.07, 3.48, 5.14]
        # E0 and V0 can be estimated from the cohesive energy, if known
        try:
            ec = ecoh[chemical_symbols[z]]
        except KeyError:
            ec = 2.5
        parameters.append(-abs(ec))  # E0
        parameters.append(0.68 * abs(ec))  # V0
        # S0
        refstate = reference_states[z]
        if refstate['symmetry'] == 'fcc':
            a = refstate['a'] 
        elif refstate['symmetry'] == 'hcp':
            a = refstate['a'] * sqrt(2)
        else: 
            # Assume BCC.  If something else the BCC guess is probably better than nothing.
            a = refstate['a'] / (0.5**0.3333)
        parameters.append(a * 0.39)
        # n0  Guesstimate from inverse atomic volume.
        parameters.append(2.97 * a**(-3))
    else:
        raise RuntimeError(f"EMT parameters not known for Z={z} and guessing not allowed.")
    if order == 'delta':
        parameters[2] = beta * parameters[0] - parameters[2]

    return parameters

if __name__ == "__main__":
    # This script is an example of how to fit EMT parameters.
    import sys
    from asap3.Tools.ParameterOptimization import ParameterOptimization
    from asap3.Internal.EMTParameters import _std_emt_params
    from asap3.Tools.OptimizationDatabase import get_data

    initparameters = {('Cu','Cu'): EMTStdParameters('Cu')}

    varparameters = {('Cu','Cu'): [True, True, True, True, True, True, False]}

    # This should be generated semi-automatically
    fitquantities = [('lattice_constant_a', 'fcc', 'Cu', get_data('Cu', 'a'), 0.005),
                     ('cohesive_energy', 'fcc', 'Cu', get_data('Cu', 'Ecoh'), 0.01),
                     ('bulk_modulus', 'fcc', 'Cu', get_data('Cu', 'B'), 0.01),
                     ('elastic_constant_C11', 'fcc', 'Cu', get_data('Cu', 'C11'), 0.01),
                     ('elastic_constant_C12', 'fcc', 'Cu', get_data('Cu', 'C12'), 0.01),
                     ('elastic_constant_C44', 'fcc', 'Cu', get_data('Cu', 'C44'), 0.01),
                     ]

    # Initial guesses?
    latticeconstants = [['fcc', 'Cu', 4.00],  
                        ['bcc', 'Cu', 4.00]]

    opt = ParameterOptimization(['Cu'], EMT2013Fit, initparameters, varparameters,
                                fitquantities, latticeconstants, debug=False)
    (optparam, optvalues) = opt.fit(log=sys.stdout)

    print("Initial parameters:", initparameters)
    print("Optimal parameters:", optparam)
    #print ""
    #print "Fitting values", fitquantities
    #print "Optimal values", optvalues




#!/usr/bin/env python
"""
A collection of tools used to implement the ASAP test suite.

Name: testtools.py

Description: A collection of tools used to make it easier to write the
ASAP test suite.

Usage: Used mainly to test if tests have failed or succeeded:

         ReportTest(name, value, expected, deviation)

       checks that the obtained value is closer to the expected value
       than deviation.  Name is the name for the test.

Example usage: Run it on the command line to test the tools themselved

Expected result: A number of successful and failed tests, self explanatory.
"""


import sys

# If this module is imported from a pytest test, then it is because the
# import statement has not been upgraded.

if 'pytest' in sys.modules:
    raise ImportError('Do not import asap3.testtools.ReportTest from pytest!')

class _Reporter:
    """Check if a test has failed, reports the result and remembers failed tests.

    A single instance, ReportTest, is created in this module and
    should be used.  It is a callable object with one additional
    method.
    
    Call with four arguments:
        name: the name of the test.
        value: the result of the test.
        expected: the expected result of the test.
        deviation: the maximally allowed deviation.

    The object has the additional method Summary(exit=0), printing a
    summary.  If the optional argument exit is non-zero, a failed test
    causes the script to exit with exit the given exit code.

    """
    def __init__(self):
        self.Clear()
        self.exitaftererror = False
        
    def Clear(self):
        self.failed = []
        self.success = 0

    def __call__(self, name, value, expected, deviation, silent=False):
        "Report the result of a test."
        if abs(value - expected) <= deviation:
            # Success
            if not silent:
                print("%s: SUCCESS" % name)
            self.success += 1
        else:
            # Failure
            error = ("%s: FAILED, got %.8g expected %.8g (max deviation: %.8g)"
                     % (name, value, expected, deviation))
            print(error)
            self.failed.append(error)
            if self.exitaftererror:
                raise AssertionError(error)

    def BoolTest(self, name, value, silent=False):
        "Report the result of a boolean test"
        if value:
            if not silent:
                 print("%s: SUCCESS" % name)
            self.success += 1
        else:
            error = (f"{name}: FAILED.")
            print(error)
            self.failed.append(error)
            if self.exitaftererror:
                raise AssertionError("Exiting after error")

    def ExitAfterError(self, ex=True):
        self.exitaftererror = ex
        
    def GetNumberOfErrors(self):
        "Return the number of failed tests."
        return len(self.failed)

    def Summary(self, exit=1):
        if self.failed:
            report =  "%d tests succeeded, %d TESTS FAILED." % (self.success,
                                                                len(self.failed))
            print(report)
            print("")
            print("Failed tests:")
            for failure in self.failed:
                print("   ", failure)
            if exit:
                # TestAll.py catches the exception below, so clean up before
                # TestAll calls the next test.
                self.Clear()  
                raise AssertionError(report)
        else:
            print("ALL TESTS SUCCEEDED (%d tests were performed)" % self.success)

# The main test object:
ReportTest = _Reporter()


if __name__ == "__main__":
    # Test the test code.
    print("Testing the test code.  5 tests should succeed, 5 should fail.")
    ReportTest("Expected success 1", 42.1, 42.1, 0.0)
    ReportTest("Expected success 2", 42.101, 42.1, 0.001)
    ReportTest("Expected success 3", 42.0995, 42.1, 0.001)
    ReportTest("Expected success 4", -0.001, -0.01, 0.01)
    ReportTest("Expected success 5", 0.001, -0.005, 0.01)
    ReportTest("Expected failure 1", 42.100001, 42.1, 0.0)
    ReportTest("Expected failure 2", 42.11, 42.1, 0.001)
    ReportTest("Expected failure 3", 42.09, 42.1, 0.001)
    ReportTest("Expected failure 4", -0.001, -0.01, 0.0001)
    ReportTest("Expected failure 5", 0.001, -0.01, 0.0001)
    print("")
    ReportTest.Summary(exit=0)
    print("")
    print("IMPORTANT: This self test succeeded if EXACTLY FIVE TESTS FAILED.")
    print("")
    if len(ReportTest.failed) == 5:
        print("This test succeeded")
    else:
        print("THIS TEST FAILED !!!")
        print("5 tests should have failed, %d tests did fail" % (len(ReportTest.failed),))
        sys.exit(1)
        



# Checks for Ticket #11

from asap3 import *
from ase.lattice.cubic import FaceCenteredCubic

from asap3.test.pytest_markers import ReportTest, serial
import pytest

@serial
def test_not_wrapping():
    'Test that atoms are not wrapped when unit cell is set.'
    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(6,6,6),
                            symbol="Cu")
    atoms.calc = EMT()
    r = atoms.get_positions()
    print("Orig position", r[-1])

    uc = atoms.get_cell()
    print(uc)
    r[-1] = 1.51*uc[2]
    atoms.set_positions(r)
    print(atoms.get_potential_energy())

    p1 = atoms.get_positions()[-1]
    print("p1:", p1)

    atoms.set_cell(uc, scale_atoms=True)
    print(atoms.get_potential_energy())
    p2  = atoms.get_positions()[-1]
    print("p2:", p2)

    atoms.set_cell(uc, scale_atoms=False)
    print(atoms.get_potential_energy())
    p3 = atoms.get_positions()[-1]
    print("p3:", p3)

    ReportTest("p2 equals p1", p2[2], p1[2], 1e-6)
    ReportTest("p3 equals p1", p3[2], p1[2], 1e-6)


@serial
@pytest.mark.parametrize('pbc', (True, False, (1,1,0)), ids=('periodic', 'free', 'mixed'))
@pytest.mark.parametrize('scale', (True, False))
def test_recalc_scaling(pbc, scale):
    "Test for correct energy calculations with varying unit cell."

    print(f"Running test with pbc={pbc} and scale_atoms={scale}")
    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                                size=(6,6,6), symbol="Cu", pbc=pbc)
    atoms.calc = EMT()
    uc = atoms.get_cell()
    atoms.get_potential_energy()
    for factor in (1.0, 1.01, 1.02, 1.1, 1.5, 1.49, 1.4, 1.4, 1.0, 0.9):
        atoms.set_cell(uc * factor, scale_atoms=scale)
        f = atoms.get_forces()
        e = atoms.get_potential_energy()
        atoms2 = Atoms(atoms)
        atoms2.calc = EMT()
        e2 = atoms2.get_potential_energy()
        f2 = atoms2.get_forces()
        name = "(factor = {:.3f}  PBC = {}  scale_atoms = {})".format(factor,
                                                                pbc, scale)
        ReportTest("Energy "+name, e, e2, 1e-6)
        maxf = max(abs(f.flat[:] - f2.flat[:]))
        ReportTest("Forces "+name, maxf, 0.0, 1e-6)



import os
import math
try:
    import cPickle as pickle
except ImportError:
    import pickle
from numpy import random
import numpy as np
import sys
from ase.optimize import FIRE, MDMin
from ase.io import write
from ase.io.trajectory import Trajectory
from ase import Atoms
from ase.utils import basestring
import subprocess
import gzip
#Changelog:
#v1.1 : Added last atoms configuration to write out for visualizing
#v1.2 : Save only counts in totals not atomwise
#v1.3 : Adscalc or EMT is determined in relax.
#v1.4 : Save SMC energy initially
#v1.5 : Add multiplicity
#v1.6 : Only relax if n_step >= n_step/2 ;  Corrected multiplicities loading.

class AtomMonteCarloData:
    version=1.6
    def sumcoordinations(self,coordinations):
        #Takes coordinations listed atomwise and returns coordinations listed 
        #by counts(coordinationnumber)
        maxcoor=15
        out = np.zeros(maxcoor, int)
        for i in range(maxcoor):
            out[i] = np.equal(coordinations,i).sum() #growing array
        return out


    def __init__(self, filename=None, atoms=None,
                 surfmove=None, symbol=None, symmetry=None,
                 latticeconstant=None, size=None, temp=None,
                 mode=None,Esmc = None, multiplicity=1,total_steps=None):
        #First find atomic number if symbol is a string, else user must have             specified Z
        if isinstance(symbol,basestring):
            self.atomic_number = data.atomic_numbers[symbol]
        else:
            self.atomic_number = symbol
        #Set attributes:
        self.filename = filename
        self.symmetry = symmetry
        self.lattice_constant = latticeconstant
        self.size = size
        self.temp = temp #Simulation Temperature really.
        self.atoms = atoms
        self.surfmove = surfmove #This holds info from class SurfaceMove from             Surface.py in ase such as coordinations, types etc.
        #Maybe:
        if not self.atoms==None:
            self.natoms = len(atoms)
        else:
            self.natoms = 0
        self.ground_state_energy = 1e10
        self.multiplicity = multiplicity

        #The lists the object shall store:
        self.lists = {}
        self.lists['energy'] = []
        self.lists['counts'] = []
        self.lists['types'] = [] #This holds the atom site type e.g. hcp 
        self.lists['coordinations']=[]
        self.lists['relaxed_energies'] = []
        self.lists['unrelaxed_energies'] =[]
        #The vacant sites:
        self.lists['vac_coordinations']=[]
        self.lists['vac_types']=[]
       
        
        #Store smc energy for future analysis of E_filter
        self.Esmc = Esmc
        
        #Store MC step number : MUST be increased in accept move:
        self.mc_step = 0
        self.total_steps = total_steps
        self.id_relax = None #used to determine useable data.
        self.n_accept = 0
        
        #Debug Variables:
        self.ninsane = 0
    
    
    
    def get(self,list2get):
        if self.lists[str(list2get)] in self.lists:
            return lists
        else:
            raise ValueError("No data named "+list2get)
        
    
    
        
    def read(self,filename=None):
        if filename is None:
            filename = self.filename  
        
        if os.path.isfile(filename) and filename.endswith(".amc.gz"):
            f = gzip.open(filename, 'rb')
        elif os.path.isfile(filename) and filename.endswith(".amc"):
            f = open(filename, "rb")
        else: #File does not exist else probably
            print(file=sys.stderr)
            sys.exit(1)
        #load pickleand store all lists:
        d = pickle.load(f)
        assert(d['filetype'] == "AtomMonteCarloData") #File check.
        assert(d['version'] == self.version)
        self.lists = d['lists']
        #store the non list properties:
        self.atomic_number = d['atomic_number']
        self.lattice_constant = d['lattice_constant']
        #self.size = d['size']
        self.temp = d['temp']
        #MaybeStore these:
        self.natoms = d['natoms']
        self.Esmc = d['Esmc']
        self.multiplicity = d['multiplicity']
        self.id_relax = d['id_relax']
        self.ninsane = d['ninsane']
        self.total_steps = d['total_steps']
        self.n_accept = d['n_accept']
    
    
       
    def write(self,filename=None):
        if filename is None:
            filename = self.filename
        if not filename.endswith('.gz'):
            filename = filename + '.gz'    
        if os.path.isfile(filename):
            os.rename(filename,filename + '.bak')
        #Data is stored as dict.
        d = {'filetype' : 'AtomMonteCarloData',
             'version' : self.version, 
             'atomic_number':          self.atomic_number,
             'symmetry': self.symmetry,
             'lattice_constant':                   self.lattice_constant,
             'temp': self.temp,
             'Esmc':self.Esmc,
             'natoms': self.natoms,
             'lists':          self.lists,
             'atomspositions':self.atoms.get_positions(),
             'multiplicity': self.multiplicity,
             'id_relax':self.id_relax, 
             'ninsane':self.ninsane,
             'total_steps':self.total_steps,
             'n_accept':self.n_accept
             }
        
        #f = open(filename,'w')
        f = gzip.open(filename, 'wb')
        #f = subprocess.Popen("gzip > "+filename, shell=True,stdin=subprocess.PIPE).stdin
        pickle.dump(d,f, protocol=-1)
        f.close()
    
    
        
    def relax(self):
        """Relax the atoms return the relaxed energy.
        This function leaves the atoms unchanged (it changes them
        but restores the positions).
        
        If a single atoms moves "too much" the relaxation is rejected.
        """
        atoms = self.atoms
        calc = atoms.calc
        specialcalc = (calc.__class__.__name__ == 'adscalc' and hasattr(calc, 'relax_calc'))
        if specialcalc:
            # Copy the atoms, and use the special calculator
            atoms = Atoms(atoms, calculator=calc.relax_calc)
        elif calc.__class__.__name__ == 'adscalc':
            calc.use_adsorbates(False)
        energy_before = atoms.get_potential_energy()
        #print "unrelaxed energy: "+str(energy)
        old_pos = atoms.get_positions()
        
        for i in range(3):
            if i:
                dyn = FIRE(atoms=atoms)
                dyn.run(fmax=0.001, steps=2000)
            else:
                dyn = MDMin(atoms=atoms)
                dyn.run(fmax=0.025, steps=2000)
        
        reten=atoms.get_potential_energy()
        
        # Sanity check. Do not use the relaxed energy if cfg. is insane.
        nblist = self.surfmove.neighbors
       
        r = atoms.get_positions() - old_pos
        
        dr = np.zeros((len(atoms),3))
        for i in range(len(atoms)):
            nblist_i = nblist[i]
            dr[i] = r[i] - r[nblist_i].sum(axis=0)/len(nblist[i])
        dr = np.sqrt( (dr*dr).sum(axis=1).max() )
        # dr is now how far the most mobile atom has moved
        # relative to its neighbors.
        # Now, identify how far it is reasonable that it has moved.
        x = old_pos[0] - old_pos[nblist[0]]
        acceptable = np.sqrt( (x*x).sum(axis=1).min() )
        assert acceptable > 0.01
        if dr > 0.5 * acceptable:
            reten = energy_before  # Reject
            self.ninsane += 1
        # Restore the atoms
        atoms.set_positions(old_pos)
        if not specialcalc and calc.__class__.__name__ == 'adscalc':
            calc.use_adsorbates(True)
        return energy_before, reten 
    
    
    
    def accept_move(self):
        e = self.atoms.get_potential_energy()
        self.lists['energy'].append(e)
        self.lists['counts'].append(1)
        self.lists['coordinations'].append(self.sumcoordinations(self.surfmove.coordinations))
        self.lists['types'].append(self.sumcoordinations(self.surfmove.types)) #See sumcoordinations
        self.lists['vac_coordinations'].append(self.sumcoordinations(self.surfmove.vacant_coordinations))
        self.lists['vac_types'].append(self.sumcoordinations(self.surfmove.vacant_types))
        
        
        if self.mc_step>=self.total_steps/2:
            
            if self.id_relax == None:
                self.id_relax = self.n_accept+1
                
            unrelax_energy, relax_energy = self.relax()
            self.lists['unrelaxed_energies'].append(unrelax_energy)
            self.lists['relaxed_energies'].append(relax_energy)
            e += relax_energy - unrelax_energy
        
            if e < self.ground_state_energy:
                #now store .amc in tmpfn[1] and 0000* in tmpfn[0]
                tmpfn = self.filename.split(".")
                traj = Trajectory(tmpfn[0]+".traj", 'w', self.atoms, master=True)
                traj.write()
                traj.close()
                #write(filename=tmpfn[0]+".traj",images=self.atoms)
                self.ground_state_energy = e #Previously unindented made an error
                
        else:
            self.lists['unrelaxed_energies'].append(np.nan)
            self.lists['relaxed_energies'].append(np.nan)

        self.mc_step += 1
        self.n_accept += 1  
    
    
        
    def reject_move(self):
        self.mc_step+=1
        l = self.lists['counts']
        if l:
            l[-1] += 1
    
    
        



import os

openkimmodel = "EMT_Asap_Standard_JacobsenStoltzeNorskov_1996_AlAgAuCuNiPdPt__MO_115316750986_001"

if 'ASAP_KIM_DIR' in os.environ:
    _d = os.path.join(os.environ['ASAP_KIM_DIR'], 'src/models')
elif 'ASAP_KIM_LIB' in os.environ:
    _d  = os.path.join(os.environ['ASAP_KIM_LIB'], 'kim-api/models')
elif 'KIM_HOME' in os.environ:
    _d  = os.path.join(os.environ['KIM_HOME'], 'lib/kim-api/models')
else:
    _d = None

if _d is not None and os.path.exists(_d):
    openkimmodels = [x for x in os.listdir(_d) if os.path.isdir(os.path.join(_d,x))]
    openkimmodels.sort()
else:
    openkimmodels = []
    
if __name__ == "__main__":
    print(openkimmodels)
    print("\n\nThis it not a test, but a module imported from a few tests.")
    
    



import numpy as np

from ase import Atoms, units
from ase.data import atomic_numbers
from asap3 import Morse, PropertyNotImplementedError
from asap3.test.pytest_markers import serial

import pytest

@serial
@pytest.mark.core
def test_notimplemented():
    # Define constants and calculator
    elements = np.array([atomic_numbers['Ru'], atomic_numbers['Ar']])
    epsilon = np.array([[5.720, 0.092], [0.092, 0.008]])
    alpha = np.array([[1.475, 2.719], [2.719, 1.472]])
    rmin = np.array([[2.110, 2.563], [2.563, 4.185]])
    rcut = rmin.max() + 6.0 / alpha.min()

    calc = Morse(elements, epsilon, alpha, rmin)
    atoms = Atoms(['Ar', 'Ar'], [[0.0, 0.0, 0.0], [rcut + 1.0, 0.0, 0.0]])
    atoms.center(vacuum=10.0)
    atoms.set_pbc(True)
    atoms.calc = calc

    # This should not fail
    energy = atoms.get_potential_energy()

    #This should fail
    with pytest.raises(PropertyNotImplementedError):
        stress = atoms.get_stress()




import asap3
import numpy as np
from ase.build import bulk
from ase.eos import EquationOfState
import numpy as np
from asap3.mpi import world
from asap3.test.pytest_markers import ReportTest, withOpenKIM, serial
import itertools

import pytest

def get_model(model):
    try:
        calc = asap3.OpenKIMcalculator(model)
    except asap3.AsapError as oops:
        if oops.args[0].startswith('Failed to initialize OpenKIM model'):
            print(f"OpenKIM model {model} not installed - skipping test.")
            calc = None
        else:
            raise
    return calc

if world.size == 1:
    sizes = [(1,1,1), (2,2,2), (5,5,5)]
else:
    sizes = [(25,25,25)]

@withOpenKIM
@pytest.mark.parametrize('size', sizes, ids=[x[0] for x in sizes])
def test_par_openKIM_Ar_Morse(cpulayout, size):
    model = 'ex_model_Ar_P_Morse'
    ismaster = world.rank == 0
    calc = get_model(model)
    if calc is not None:
        if ismaster:
            atoms = bulk('Ar', 'fcc', 5.26).repeat(size)
            print(atoms.cell)
        else:
            atoms = None
        if world.size > 1:
            atoms = asap3.MakeParallelAtoms(atoms, cpulayout)
        # calc = asap3.OpenKIMcalculator(model)   # Reuse of OpenKIM objects not yet supported.
        atoms.calc = calc

        e = atoms.get_potential_energy()/atoms.get_global_number_of_atoms()
        print("Potential energy:", e)
        ReportTest(f"Potential energy {str(size)}", e, -0.092798, 1e-5)

        eq_cell = atoms.get_cell()
        scales = np.linspace(0.97, 1.03, 7)
        energies = []
        volumes = []
        for s in scales:
            atoms.set_cell(s * eq_cell, scale_atoms=True)
            energies.append(atoms.get_potential_energy())
            volumes.append(atoms.get_volume())
        eos = EquationOfState(volumes, energies)
        v0, e0, B = eos.fit()
        v_cell = v0 * 4 / atoms.get_global_number_of_atoms()
        a0 = v_cell**(1/3)
        print("Lattice constant:", a0)
        print("Bulk modulus:", B)
        ReportTest(f"Lattice constant {str(size)}", a0, 5.2539, 1e-4)


@withOpenKIM
@pytest.mark.parametrize('size', sizes, ids=[x[0] for x in sizes])
def test_par_openKIM_multi_nbl(cpulayout, size):
    model = 'ex_model_Ar_SLJ_MultiCutoff'   # Requires multiple neighbor lists.
    ismaster = world.rank == 0
    calc = get_model(model)
    if calc is not None:
        if ismaster:
            atoms = bulk('Ar', 'fcc', 5.26).repeat(size)
        else:
            atoms = None
        if world.size > 1:
            atoms = asap3.MakeParallelAtoms(atoms, cpulayout)
        # calc = OpenKIMcalculator(model)   # Reuse of OpenKIM objects not yet supported.
        atoms.calc = calc

        e = atoms.get_potential_energy()/atoms.get_global_number_of_atoms()
        if ismaster:
            print("Potential energy:", e)
        ReportTest(f"Potential energy {str(size)}", e, -0.0104, 1e-6)

        eq_cell = atoms.get_cell()
        scales = np.linspace(0.99, 1.01, 7)
        energies = []
        volumes = []
        for s in scales:
            atoms.set_cell(s * eq_cell, scale_atoms=True)
            energies.append(atoms.get_potential_energy())
            volumes.append(atoms.get_volume())
        eos = EquationOfState(volumes, energies)
        v0, e0, B = eos.fit()
        v_cell = v0 * 4 / atoms.get_global_number_of_atoms()
        a0 = v_cell**(1/3)
        if ismaster:
            print("Lattice constant:", a0)
            print("Bulk modulus:", B)
        ReportTest(f"Lattice constant {str(size)}", a0, 5.26, 1e-5)


@withOpenKIM
@serial
# @pytest.mark.parametrize('reuse', (False, True))
def test_openkim_LJ_modify(reuse=False):
    # Reuseing does not work due to a neighborlist problem - should be fixed?
    model = 'LennardJones612_UniversalShifted__MO_959249795837_003'
    calc = get_model(model)
    p = calc.parameters
    # Get some parameters
    epsilon_He = p.get_parameter('epsilons', 'ut')[1,1]
    sigma_He = p.get_parameter('sigmas', 'ut')[1,1]
    # Set a parameter
    p['shift'] = 0

    helium = bulk("He", 'fcc', np.sqrt(2) * 2**(1/6) * sigma_He).repeat((5,5,5))
    helium.calc = calc

    e = helium.get_potential_energy() / len(helium)
    ReportTest("Helium energy", e, -8.24646294 * epsilon_He, 1e-6)

    if not reuse:
        calc2 = asap3.OpenKIMcalculator(model)
    else:
        calc2 = calc
    eps = calc2.parameters.get_parameter('epsilons', 'ut')
    eps[1,1] = 2 * epsilon_He
    calc2.parameters.set_parameter('epsilons', eps, 'ut')
    calc2.parameters['shift'] = 0

    if not reuse:
        helium.calc = calc2
    e2 = helium.get_potential_energy() / len(helium)

    ReportTest("Modified Helium energy", e2, 2*e, 1e-6)

emt_openkimmodel = 'EMT_Asap_Standard_JacobsenStoltzeNorskov_1996_%s__%s'

emt_shortkimid = {
    'Ag': 'MO_303974873468_001',
    'Al': 'MO_623376124862_001',
    'Au': 'MO_017524376569_001',
    'Cu': 'MO_396616545191_001',
    'Ni': 'MO_108408461881_001',
    'Pd': 'MO_066802556726_001',
    'Pt': 'MO_637493005914_001',
    }

if asap3.OpenKIMsupported:  
    elements = ('Al', 'Ag', 'Au', 'Cu', 'Ni', 'Pd', 'Pt')
else:
    elements = ()
pbc_list = ((1,1,1), (0,0,0), (0,0,1))
size_list = ((10, 10, 10), (20, 5, 2), (2, 2, 2), (1, 1, 1))

@withOpenKIM
@serial
@pytest.mark.parametrize('element', elements)
@pytest.mark.parametrize('pbc', pbc_list)
@pytest.mark.parametrize('size', size_list)
def test_openkim_emt(element, pbc, size):
    step = 500
    for i in (0, 1):
        #pbc = next(pbc_list)
        #size = next(size_list)
        txt = ("%s=%i%i%i-%i-%i-%i " % ((element,) + pbc + size))
        # Test that EMT reimported through OpenKIM gives the right results.
        atoms_kim = bulk(element).repeat(size)
        #atoms_kim = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
        #                    size=(30, 30, 30),
        #                    symbol="Cu")
        natoms = len(atoms_kim)
        atoms_kim.set_pbc(pbc)
        r = atoms_kim.get_positions()
        r.flat[:] += 0.1 * np.sin(np.arange(3*natoms))
        atoms_kim.set_positions(r)
        atoms_emt = atoms_kim.copy()
        try:
            kim = asap3.OpenKIMcalculator(emt_openkimmodel % (element, emt_shortkimid[element]))
        except asap3.AsapError as oops:
           if oops.args[0].startswith('Failed to initialize OpenKIM model'):
               print("OpenKIM model {} not installed - skipping test.".format(
                   emt_openkimmodel % (element, emt_shortkimid[element])))
               continue
           else:
               raise   # Something else went wrong.
        emt = asap3.EMT()
        emt.set_subtractE0(False)
        atoms_kim.calc = kim
        atoms_emt.calc = emt
        ek = atoms_kim.get_potential_energy()
        ee = atoms_emt.get_potential_energy()
        ReportTest(txt+"Total energy", ek, ee, 1e-8)
        ek = atoms_kim.get_potential_energies()
        ee = atoms_emt.get_potential_energies()
        for i in range(0, natoms, step):
            ReportTest(txt+"Energy of atom %i" % (i,), ek[i], ee[i], 1e-8)
        fk = atoms_kim.get_forces()
        fe = atoms_emt.get_forces()
        n = 0
        for i in range(0, natoms, step):
            n = (n + 1) % 3
            ReportTest(txt+"Force(%i) of atom %i" % (n, i), fk[i, n], fe[i, n], 1e-8)
        sk = atoms_kim.get_stress(include_ideal_gas=True)
        se = atoms_emt.get_stress(include_ideal_gas=True)
        for i in range(6):
            ReportTest(txt+"Stress(%i)" % (i,), sk[i], se[i], 1e-8)
        sk = atoms_kim.get_stresses(include_ideal_gas=True)
        se = atoms_emt.get_stresses(include_ideal_gas=True)
        for i in range(0, natoms, step):
            n = (n + 1) % 6
            # Volume per atom is not defined the same way: greater tolerance needed
            ReportTest(txt+"Stress(%i) of atom %i" % (n, i), sk[i, n], se[i, n], 1.5e-3)
    



"""Python definition of EMT parameters."""

from asap3.Internal.BuiltinPotentials import EMTParameters
import copy
import numpy as np
import ase

bohr = 0.5291772;  # Same number of digits as in EMTDefaultParameters
bohr3 = bohr * bohr * bohr

_std_emt_params = {
    13: {"name" :  "Al",
         "E0": -3.280,
         "S0": 3.000 * bohr,
         "V0": 1.493,
         "eta2": 1.240 / bohr,
         "kappa": 2.000 / bohr,
         "lambda": 1.169 / bohr,
         "n0": 0.007 / bohr3},
    29: {"name":  "Cu",
         "E0": -3.510,
         "S0": 2.67 * bohr,
         "V0": 2.476,
         "eta2": 1.652 / bohr,
         "kappa": 2.74 / bohr,
         "lambda": 1.906 / bohr,
         "n0": 0.0091 / bohr3},
    47: {"name": "Ag",
         "E0": -2.96,
         "S0": 3.01 * bohr,
         "V0": 2.132,
         "eta2": 1.652 / bohr,
         "kappa": 2.790 / bohr,
         "lambda": 1.892 / bohr,
         "n0": 0.00547 / bohr3},
    79: {"name":  "Au",
         "E0": -3.80,
         "S0": 3.00 * bohr,
         "V0": 2.321,
         "eta2": 1.674 / bohr,
         "kappa": 2.873 / bohr,
         "lambda": 2.182 / bohr,
         "n0": 0.00703 / bohr3},
    28: {"name":  "Ni",
         "E0": -4.44 ,
         "S0": 2.60  * bohr,
         "V0": 3.673,
         "eta2": 1.669 / bohr,
         "kappa": 2.757 / bohr,
         "lambda": 1.948 / bohr,
         "n0": 0.0103 / bohr3},
    46: {"name":  "Pd",
         "E0": -3.90 ,
         "S0": 2.87 * bohr,
         "V0": 2.773,
         "eta2": 1.818 / bohr,
         "kappa": 3.107 / bohr,
         "lambda": 2.155 / bohr,
         "n0": 0.00688 / bohr3},
    78: {"name":  "Pt",
         "E0": -5.85 ,
         "S0": 2.90 * bohr,
         "V0": 4.067,
         "eta2": 1.812 / bohr,
         "kappa": 3.145 / bohr,
         "lambda": 2.192 / bohr,
         "n0": 0.00802 / bohr3},
    }

class EMTStandardParameters(EMTParameters):
    """Standard EMT parameters.

    This is a Python implementation of EMTDefaultParameters,
    intended for use as a base class.
    """
    _defaultparameters = copy.deepcopy(_std_emt_params)
    beta = 1.809;  # Preserve same rounding as in ARTwork.
    
    def __init__(self):
        # Make a copy of parameters so we do not risk modifying the
        # orgininal ones in derived clases.
        EMTParameters.__init__(self)
        self.parameters = copy.deepcopy(self._defaultparameters)
        self.actual_params = []
        self.gammas_calculated = False
        self.shellpop = [12, 6, 24, 12, 24]
        self.shelldist = np.sqrt(np.arange(1,6))
        self.shell_cutoff = 3
        self.shell_gamma = 3
        self.listcutofffactor = 1.04500185048
        
    def get_maximal_cutoff(self):
        """The maximal cutoff value.  Defined before the elements is known."""
        maxS0 = 0.0
        for p in self.parameters.values():
            S0 = p['S0']
            if S0 > maxS0:
                maxS0 = S0
        cutoff = 0.5 * maxS0 * self.beta * (np.sqrt(self.shell_cutoff) +
                                            np.sqrt(self.shell_cutoff + 1))
        return cutoff * self.listcutofffactor
    
    def get_parameters(self, element):
        assert self.gammas_calculated == False
        p = self.parameters[element]
        p["Z"] = element
        p["mass"] = ase.data.atomic_masses[element]
        p["name"] = ase.data.chemical_symbols[element]
        self.actual_params.append(p)
        return p

    def get_gammas_etc(self):
        assert self.gammas_calculated == False
        self.calc_cutoff()
        self.calc_gammas()
        self.calc_chi()
        self.gammas_calculated = True
        return (self.cutoff, self.gammas, self.chi)

    def calc_cutoff(self):
        maxS = 0.0
        for p in self.actual_params:
            if p["S0"] > maxS:
                maxS = p["S0"]
        cutoff = 0.5 * maxS * self.beta * (np.sqrt(self.shell_cutoff) +
                                           np.sqrt(self.shell_cutoff + 1))
        r = (cutoff * 2.0 * np.sqrt(self.shell_cutoff + 1)
             / (np.sqrt(self.shell_cutoff) + np.sqrt(self.shell_cutoff + 1)))
        cutslope = np.log(9999.0) / (r - cutoff);
        self.cutoff = (cutoff, cutslope, self.listcutofffactor)

    def calc_gammas(self):
        self.gammas = []
        for p in self.actual_params:
            gamma1 = gamma2 = 0.0
            for i in range(self.shell_gamma):
                d = self.shelldist[i] * self.beta * p["S0"]
                w = 1. / (1. + np.exp(self.cutoff[1] * (d - self.cutoff[0])))
                gamma1 += w * self.shellpop[i] * np.exp(-d * p["eta2"])
                gamma2 += w * self.shellpop[i] * np.exp(-d * p["kappa"]
                                                        / self.beta)
            gamma1 /= 12 * np.exp(-self.beta * p["S0"] *
                                                p["eta2"])
            gamma2 /= 12 * np.exp(-p["S0"] * p["kappa"])
            self.gammas.append((gamma1, gamma2))

    def calc_chi(self):
        n = len(self.actual_params)
        self.chi = np.zeros((n,n), float)
        for i in range(n):
            for j in range(n):
                self.chi[i,j] = (self.actual_params[j]["n0"]
                                 / self.actual_params[i]["n0"] )
        
class EMThcpParameters(EMTStandardParameters):
    """EMT parameters for elements in HCP structure.

    Currenlty, only Ruthenium is supported.
    """
    _hcp_parameters = {44: {"name": "Ru",
                            "E0": -6.350455,
                            "V0": 12.305218,
                            "S0": 1.498407,
                            "eta2": 3.945242,
                            "kappa": 6.561249,
                            "lambda": 6.701546,
                            "n0": 3.917210,
                            }
                       }
    
    def __init__(self):
        self.parameters = copy.deepcopy(self._hcp_parameters)
        self.actual_params = []
        self.gammas_calculated = False
        self.shellpop = (6, 6, 6, 2, 6, 12, 12, 6)
        c = 2*np.sqrt(2.0/3.0)
        self.shelldist = (1.0,
                          np.sqrt(c*c/4.0 + 1/3.0),
                          np.sqrt(c*c/4.0 + 4.0/3.0),
                          c,
                          np.sqrt(3.0),
                          np.sqrt(7.0/3.0 + c*c/4.0),
                          np.sqrt(1+c*c),
                          2.0)
        self.shell_gamma = 8
        self.shell_cutoff = 3  # Not really a shell number, but refers to FCC
        self.listcutofffactor = 1.04500185048
        
    def calc_cutoff(self):
        maxS = 0.0
        for p in self.actual_params:
            if p["S0"] > maxS:
                maxS = p["S0"]
        cutoff = maxS * self.beta * 1.824;
        r = (cutoff * 2.0 * np.sqrt(self.shell_cutoff + 1)
             / (np.sqrt(self.shell_cutoff) + np.sqrt(self.shell_cutoff + 1)))
        cutslope = np.log(9999.0) / (r - cutoff);
        listcutoffdistance = np.log(9999.0) / cutslope + cutoff;
        if listcutoffdistance > cutoff * self.listcutofffactor :
            listcutofffactor = listcutoffdistance / cutoff
        else:
            listcutofffactor = self.listcutofffactor
        self.cutoff = (cutoff, cutslope, listcutofffactor)

class EMTMetalGlassParameters(EMTStandardParameters):
    """EMT parameters for CuMg and CuZr metallic glasses.

    IMPORTANT: These copper parameters are different from the usual Cu
    parameters, and should ONLY be used to simulate metallic glasses.
    """
    _defaultparameters = {12: {"name": "Mg",
                               "E0":  -1.487000,
                               "eta2": 2.541137,
                               "kappa": 4.435425,
                               "lambda": 3.292725,
                               "n0": 0.035544,
                               "S0": 1.766399,
                               "V0": 2.229870},
                          29: {"name": "Cu",
                               "E0": -3.510000,
                               "eta2": 3.039871,
                               "kappa": 4.943848,
                               "lambda": 3.693666,
                               "n0": 0.063738,
                               "S0": 2.67 * bohr,  # Unchanged.
                               "V0": 1.993953},
                          40: {"name": "Zr",
                               "E0": -6.300000,
                               "eta2": 2.282059,
                               "kappa": 3.911040,
                               "lambda": 2.247348,
                               "n0": 0.031831,
                               "S0": 1.785939,
                               "V0": 2.320809}
                          }



del bohr, bohr3

                               
    



"""Testing the parallel PrimiPlotter."""

from asap3 import *
from asap3.md.verlet import VelocityVerlet
from numpy import *
from pickle import *
#from Asap.Trajectories.NetCDFTrajectory import *
import sys, os, time
from asap3.testtools import *
from asap3.mpi import world
from ase.build import bulk
from ase import units
from asap3.visualize.primiplotter  import *
import shutil

debug = 0
if debug == 1:
    DebugOutput("makeverlet%d.log", nomaster=True)
elif debug == 2:
    time.sleep(world.rank)
    print("PID:", os.getpid())
    time.sleep(20)

print_version(1)
set_verbose(0)

ismaster = world.rank == 0
isparallel = world.size != 1
if world.size == 1:
    cpulayout = None
elif world.size == 2:
    cpulayout = [2,1,1]
elif world.size == 3:
    cpulayout = [1,3,1]
elif world.size == 4:
    cpulayout = [2,1,2]


gs_installed = shutil.which('gs')
if not gs_installed:
    print("Skipping test, as ghostscript (gs) is not installed.")
else:
    if isparallel:
        print("RUNNING PARALLEL VERSION OF TEST SCRIPT")
    else:
        print("RUNNING SERIAL VERSION OF TEST SCRIPT")

    if ismaster:
        atoms = bulk('Cu', cubic=True).repeat((8,8,8))
        atoms.set_pbc(False)
        atoms.center(vacuum=5.0)
        if cpulayout:
            atoms = atoms.repeat(cpulayout)
        dx = 0.1 * sin(arange(3*len(atoms))/10.0)
        dx.shape = (-1,3)
        atoms.set_positions(atoms.get_positions() + dx)
        del dx
        out = sys.stderr
    else:
        atoms = None
        out = open("/dev/null", "w")

    if isparallel:
        atoms = MakeParallelAtoms(atoms, cpulayout)
        nTotalAtoms = atoms.get_global_number_of_atoms()
    else:
        nTotalAtoms = len(atoms)

    #report()

    print("Setting potential")
    atoms.calc = EMT()

    dyn = VelocityVerlet(atoms, 5 * units.fs)

    print("Number of atoms:", nTotalAtoms)

    epot = atoms.get_potential_energy() / nTotalAtoms
    ekin = atoms.get_kinetic_energy() / nTotalAtoms
    etotallist = [epot+ekin]
    ekinlist = [ekin]

    #report()

    if ismaster:
        print("\nE_pot = {:<12.5f}  E_kin = {:<12.5f}  E_tot = {:<12.5f}".format(epot, ekin,
                                                                     epot+ekin))
    dyn.attach(MDLogger(dyn, atoms, "-", peratom=True), interval=10)

    dyn.run(100)
    e0 = (atoms.get_potential_energy() + atoms.get_kinetic_energy()) / nTotalAtoms

    if isparallel:
        plotter = ParallelPrimiPlotter(atoms)
    else:
        plotter = PrimiPlotter(atoms)
    plotter.set_output(PngFile("pplottest"))  # Save as PNG files
    plotter.set_output(PostScriptFile("pplottest"))  # Save as PS files
    dyn.attach(plotter.plot, interval=250)

    filenames =[ f'pplottest{i:04d}.{t}' for i in (0, 1) for t in ('png', 'ps')]
    if ismaster:
        for f in filenames:
            if os.path.exists(f):
                print("Removing", f)
                os.unlink(f)

    dyn.run(500)

    e = (atoms.get_potential_energy() + atoms.get_kinetic_energy()) / nTotalAtoms

    ReportTest("Energy is conserved", e, e0, 1e-4)

    # Check the output files
    if ismaster:
        for f in filenames:
            ReportTest(f"File {f} exists", os.path.exists(f), 1, 0)
            if os.path.exists(f):
                ReportTest(f"File {f} is not empty", os.path.getsize(f) != 0, 1, 0)
                print("Removing", f)
                os.unlink(f)

    ReportTest.Summary(0)




"""Writes the version.c file containing version information.

Usage:  call from makefile like this:
  python recordversion.py '$(VERSION)' 'edition' '$(CXX)' '$(CXXFLAGS)' > version.cpp
where edition specifies serial or parallel editions.
"""

from time import *
from sys import argv
import os

contents = '''
// version.cpp
//
// This file is AUTOMATICALLY GENERATED BY recordversion.py.  DO NOT EDIT!
//

#include "AsapNamespace.h"
#include <string>
using std::string;

namespace ASAPSPACE {

string GetVersion()
{
  return ("ASAP version %s %s, compiled " \
          __DATE__ " " __TIME__ \
          " on %s using '%s'");
}

string GetShortVersion()
{
  return "%s";
}

} // end namespace
'''

if __name__ == '__main__':
    assert(len(argv) == 5)
    version = argv[1]
    edition = argv[2]
    if argv[4] != "":
        compiler = argv[3] + " " + argv[4]
    else:
        compiler = argv[3]
    #when = strftime("%Y-%m-%d %H:%M", localtime(time()))
    try:
        host = os.uname()[1]
    except:
        host = 'unknown'


    print(contents % (version, edition, host, compiler, version))



"""asap3.analysis.cutcluster - cut out clusters of atoms based on certain conditions."""


from asap3.analysis.findcluster import findclusters
import ase.io
import os
import sys
import numpy as np

def cut_clusters_by_size(atoms, tag, cutoff, size, folder, environment=0,
                         environ_exclude=[1,]):
    """Find clusters of specific size, save to files.

    Find connected clusters (using asap3.analysis.findcluster.findclusters),
    pick the ones with a given size, and save them to a folder as Trajectory
    files.

    Parameters:

    atoms: The atoms to be analyzed.

    tag: The tag identifying interesting atoms.  atoms.get_tags() is
        used to read the tags from the atoms, and only atoms matching
        'tag' are considered in the clustering.  Set tag=None to
        consider all atoms (for a dense sample this gives a single
        cluster spanning the entire system).

    cutoff: The cutoff distance used to identify if two atoms are
        neighbors.

    size: The cluster size(s) picked out.  Either an integer, or a
        sequence of integers

    folder: The folder where the output is saved.

    environment=0.0: If a nonzero number is given, any atom within that
        range will be included in the cluster, unless the tag is in
        environ_exclude.

    environ_exclude=[1,]:  Atoms with these tags are not included in the
        environment.

    """
    _, clustermap = findclusters(atoms, tag, cutoff, returnclusters=True)
    if size not in clustermap:
        raise RuntimeError(f"No clusters of size {size} found.")
    os.mkdir(folder)
    for i, cluster in enumerate(clustermap[size]):
        cl_atoms = atoms[cluster]
        cl_atoms.set_pbc(False)
        cl_atoms.center(vacuum=1.0)
        assert len(cl_atoms) == size
        filename = os.path.join(folder, f"cluster_{size}_{i}.traj")
        ase.io.write(filename, cl_atoms)
        print("Wrote", filename)
        
if __name__ == '__main__':
    if len(sys.argv) != 5:
        print("Usage:")
        print(" python -m asap3.analysis.cutcluster infile.bundle outfolder TAG SIZE")
        sys.exit(1)
    infile = sys.argv[1]
    folder = sys.argv[2]
    tag = int(sys.argv[3])
    size = int(sys.argv[4])
    cutoff = 3.615 * (1/np.sqrt(2) + 1) / 2
    atoms = ase.io.read(infile)
    cut_clusters_by_size(atoms, tag, cutoff, size, folder)
    
    



"""A radial distribution function observer."""

import numpy
import asap3
from asap3.Internal.Subject import Subject
from asap3.Internal.ListOfElements import ListOfElements
from asap3.mpi import world
try:
    import cPickle as pickle
except ImportError:
    import pickle
import sys
import numbers

class RadialDistributionFunction(Subject):
    def __init__(self, atoms, rMax, nBins, groups=None, interval=1, average=1,
                 autoclear=False, verbose=False):
        """Create a RadialDistributionFunction observer.

        Arguments:
        
        atoms: The atoms being observed.
        
        rMax: The maximum distance in the RDF.
        
        nBins: The number of bins in the histograms.
        
        groups (optional): A non-negative integer per atom, used to
        classify atoms into groups, the RDF is calculated for each
        group.

        interval (optional, default=1): How often should the RDF be
        calculated.  DEPRECATED: Use interval argument when attaching
        to dynamics instead.

        average (optional, default=1): How many times should the RDF
        be calculated and averaged before any observer of this object
        is notified and/or the RDF is saved to a file.

        autoclear (optional, default=False): Should the RDF be cleared
        after the RDF has been processed by observers and/or written
        to a file?  The default is to continue to accumulate data.
        """

        if not hasattr(atoms, "get_global_number_of_atoms"):
            # ASE 3.18.0 (and older) compatibility)
            atoms.get_global_number_of_atoms = atoms.get_number_of_atoms

        Subject.__init__(self)
        self.atoms = atoms
        self.natoms = atoms.get_global_number_of_atoms()
        self.globalrdf = None
        self.rMax = rMax * 1.0  # Guard against integer divisions.
        self.nBins = nBins
        self.dr = self.rMax / nBins
        self.interval = interval
        self.average = average
        self.autoclear = autoclear
        self.verbose = verbose
        self.autosave = False
        self.n1 = 0  # Counter associated with interval
        self.n2 = 0  # Counter associated with average
        self.clearnext = False
        self.countRDF = 0
        self.listOfElements = ListOfElements(atoms)
        
        if groups is None:
            self.groups = numpy.zeros(len(atoms), numpy.int32)
            self.ngroups = 1
        else:
            if groups.shape != (len(atoms),):
                raise ValueError("groups must be an integer per atom")
            if min(groups) < 0 or max(groups) >= len(atoms):
                raise ValueError("groups array is unreasonable")
            self.groups = groups.astype(numpy.int32)
            self.ngroups = max(self.groups) + 1

    def update(self, atoms=None):
        """Calculate the RDF of the atoms.

        Make an RDF calculation now (or if interval=n was specified,
        every n time this method is called.

        If average = 1, then any observers of this object is notified.

        If average > 1, the calculated RDFs are accumulated, and once
        a sufficient number of RDFs have been accumulated, the
        observers are notified and the averaged RDFs are ready.
        """
        if atoms is not None:
            self.atoms = atoms

        self.n1 += 1
        if self.n1 >= self.interval:
            if self.clearnext:
                self.clear()
                self.clearnext = False
            self.n1 = 0
            self.do_rdf()
            self.n2 += 1
            if self.n2 >= self.average:
                self.clearnext = self.autoclear
                self.call_observers()  # I have data ready
                if self.autosave:
                    self.save()
                self.n2 = 0

    def do_rdf(self):
        """Do the actual RDF calculation.  Do not call directly."""
        if self.verbose:
            print("Calculating RDF", file=sys.stderr)
        assert self.natoms == self.atoms.get_global_number_of_atoms()
        
        # Check if we need to replace the atoms with a copy.
        atoms = self.atoms
        if world.size > 1:
            try:
                calc = atoms.calc
            except AttributeError:
                calc = None
            if calc is not None:
                cutoff = calc.get_cutoff()
                if cutoff < self.rMax:
                    if self.verbose:
                        print("Cutoff too long for calculator - copying atoms", file=sys.stderr)
                    atoms = asap3.MakeParallelAtoms(atoms, atoms.nCells)
            
        grdf, rdfs, cnts = asap3._asap.RawRDF(atoms, self.rMax, self.nBins,
                                              self.groups, self.ngroups,
                                              self.listOfElements)
        # Global communication in parallel simulations
        if world.size > 1:
            if self.verbose:
                print("Collecting data", file=sys.stderr)
            world.sum(grdf)
            for i in range(len(rdfs)):
                keys = rdfs[i].keys()
                for key in sorted(keys):
                    world.sum(rdfs[i][key])
            for cnt in cnts:
                keys = cnt.keys()
                for key in sorted(keys):
                    cnt[key] = world.sum(cnt[key])
                
        # Sanity check on counts
        c = 0
        for cnt in cnts: # Loop over groups
            for t in cnt.keys(): # Loop over elements
                c += cnt[t]
        assert c == self.natoms
        # Now store the data
        if self.globalrdf is None:
            self.globalrdf = grdf
            self.rdfs = rdfs
            self.atomcounts = cnts
            # Ensure that all elements are represented
            for i in range(len(rdfs)): # Loop over groups
                for e in self.listOfElements: # Loop over elements
                    if e not in self.atomcounts[i]:
                        self.atomcounts[i][e] = 0
        else:
            self.globalrdf += grdf
            for i in range(len(rdfs)): # Loop over groups
                for t in rdfs[i].keys(): # Loop over pairs
                    self.rdfs[i][t] += rdfs[i][t]
                for t in cnts[i].keys(): # Loop over elements
                    self.atomcounts[i][t] += cnts[i][t]
        self.countRDF += 1
        self.volume = atoms.get_volume()
        if self.verbose:
            print("RDF done!", file=sys.stderr)

    def clear(self):
        """Clear the accumulated RDFs.  Called automatically."""
        if self.verbose:
            print("Clearing RDF", file=sys.stderr)
        self.globalrdf = self.rdfs = self.atomcounts = None
        self.countRDF = 0
        
    def get_rdf(self, groups=None, elements=None, normalize='volume'):
        """Get an RDF.

        Arguments (both optional):
        
        groups: Only get the RDF for atoms in these groups. Can either be an
        integer or a list integer refering to the groups (default: all groups)
        
        elements: Get the partial RDF for one element given as an atomic number
        or two elements givenas a tuple of two atomic numbers (a, b). The
        returned RDF either tells how many neighbors an atom has or how many b
        neighbors an a atom has respectively.

        If get_rdf is called on a newly created
        RadialDistributionFunction object, and this object has been
        created without specifying interval and average parameters, it
        is assumed that the user is not using the object as an
        observer, but wants an immediate calculation of the RDF.  In
        that case, calling get_rdf triggers the calculation of the
        RDFs.  In all other cases previously stored RDFs are returned.
        """
        if self.globalrdf is None and self.interval == 1 and self.average == 1:
            self.update()
            
        if groups is None and elements is None:
            # Return global RDF
            return self.normalize(self.globalrdf,
                                  self.countRDF * self.natoms,
                                  normalize)

        # Sum over selected groups
        if groups is None:
            groups = range(len(self.rdfs))
        else:
            if not isinstance(groups, list):
                groups = [groups]

        rdfs = self.rdfs[groups[0]].copy()  # Shallow copy
        atomcounts = self.atomcounts[groups[0]].copy()  # Shallow copy
        for i in groups[1:]: # Loop over groups
            if i >= len(self.rdfs):
                continue
            for t in self.rdfs[i].keys(): # Loop over pairs
                rdfs[t] += self.rdfs[i][t]
            for t in self.atomcounts[i].keys(): # Loop over elements
                atomcounts[t] += self.atomcounts[i][t]

        # Sum over selected element pairs
        elementpairs = elements
        if elementpairs is None:
            elements = self.listOfElements
            elementpairs = rdfs.keys()
        else:
            if isinstance(elementpairs, tuple):
                elements = [elementpairs[0]]
                elementpairs = [elementpairs]
            elif isinstance(elementpairs, numbers.Integral):
                elements = [elementpairs]
                elementpairs = []
                for e1 in self.listOfElements:
                    elementpairs.append((elements[0], e1))
            else:
                raise ValueError('Elements must be either an interger or ' +
                                 'a tuple of two integers.')

        rdf = None
        for pair in elementpairs:
            if rdf is None:
                rdf = numpy.array(rdfs[pair])
            else:
                rdf += rdfs[pair]

        # The atomcounts should be summed over the selected element(s)!
        atomcount = None
        for e in elements:
            if atomcount is None:
                atomcount = atomcounts[e]
            else:
                atomcount += atomcounts[e]
        if self.verbose:
            print('Number of selected atoms:', atomcount, file=sys.stderr)

        return self.normalize(rdf, atomcount, normalize)

    def normalize(self, rdf, ncount, type):
        """Normalize the raw RDF returned by the C++ module."""
        if type == 'volume':
            factor = (4 * numpy.pi / 3.0) * ncount * self.natoms / self.volume
            r = numpy.arange(self.nBins) * self.dr
            r3low = r * r * r
            r += self.dr
            r3high = r * r * r
            normalization = factor * (r3high - r3low)
        elif type == 'atoms':
            normalization = 1.0 * ncount
        else:
            normalization = 1.0
        return rdf / normalization
    
    # Saving and loading RDF's
    def output_file(self, prefix):
        """Give the file prefix for saving RDFs, and turn on saving."""
        self.autosave = True
        self.savefilecounter = 0
        if "%" in prefix:
            # Assume the user knows what (s)he is doing
            self.filenames = prefix
        else:
            self.filenames = prefix + "%04d.rdf"

    def save(self, filename=None):
        if self.verbose:
            print("Saving RDF", file=sys.stderr)
        if filename == None:
            filename = (self.filenames % (self.savefilecounter,))
            self.savefilecounter += 1
        data = {"globalrdf": self.globalrdf,
                "rdfs" : self.rdfs,
                "atomcounts": self.atomcounts,
                "countRDF": self.countRDF,
                "listOfElements": self.listOfElements,
                "rMax": self.rMax,
                "dr": self.dr,
                "nBins": self.nBins,
                "natoms": self.natoms,
                "volume": self.volume}
        f = open(filename, "wb")
        pickle.dump(data, f, 2) # Protocol 2 (efficient bin from Python 2.3)
        f.close()
        
    @classmethod
    def load(cls, filename):
        import new  # Broken in Python 3
        f = open(filename, "rb")
        data = pickle.load(f)
        f.close()
        obj = new.instance(cls, data)
        obj.verbose = False
        return obj
    



"Test the RawRadialDistributionFunction function."

from asap3 import EMT
from asap3 import _asap
from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.compounds import NaCl
from asap3.Internal.ListOfElements import ListOfElements
from asap3.test.pytest_markers import ReportTest, serial
from numpy import sqrt, zeros, int32

import pytest

# testtypes: latticeconst, maxRDF, bins, useEMT
testtypes = ((3.6, 6.001, 100, False),
             (3.7, 6.001, 100, True),
             (3.6, 15.001, 100, False),
             (3.65, 15.001, 100, True))

@serial
@pytest.mark.core
@pytest.mark.parametrize('latconst,maxrdf,nbins,withemt', testtypes)
def test_raw_rdf_fcc(latconst, maxrdf, nbins, withemt):

    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                              symbol="Cu", size=(10,10,10),
                              latticeconstant=latconst, debug=0)
    natoms = len(atoms)
    ReportTest("Number of atoms", natoms, 4000, 0)

    if withemt:
        atoms.calc = EMT()
        print(atoms.get_potential_energy())

    result = _asap.RawRDF(atoms, maxrdf, nbins, zeros(len(atoms), int32), 1,
                          ListOfElements(atoms))
    z = atoms.get_atomic_numbers()[0]

    globalrdf, rdfdict, countdict = result

    print(globalrdf)

    ReportTest("Local and global RDF are identical",
               min( globalrdf == rdfdict[0][(z,z)]), 1, 0)
    ReportTest("Atoms are counted correctly", countdict[0][z], natoms, 0)
    
    shellpop = [12, 6, 24, 12, 24, -1]
    shell = [sqrt(i+1.0)/sqrt(2.0) for i in range(6)]
    print(shell)
    print(shellpop)
    n = 0

    dr = maxrdf/nbins

    for i in range(nbins):
        if (i+1)*dr >= shell[n] * latconst:
            if shellpop[n] == -1:
                print("Reached the end of the test data")
                break
            ReportTest(("Shell %d (%d)" % (n+1, i)), globalrdf[i],
                       natoms*shellpop[n], 0)
            n += 1
        else:
            ReportTest(("Between shells (%d)" % (i,)), globalrdf[i], 0, 0)


@serial
@pytest.mark.core
def test_raw_rdf_partial():
    latconst, maxrdf, nbins, withemt = 3.6, 6.001, 100, False

    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], symbol="Cu",
                            size=(10,10,10), latticeconstant=latconst, debug=0)
    natoms = len(atoms)
    z = atoms.get_atomic_numbers()
    z[100] = 47
    atoms.set_atomic_numbers(z)

    ReportTest("Number of atoms", natoms, 4000, 0)

    if withemt:
        atoms.calc = EMT()
        print(atoms.get_potential_energy())

    result = _asap.RawRDF(atoms, maxrdf, nbins, zeros(len(atoms), int32), 1,
                        ListOfElements(atoms))

    globalrdf, rdfdict, countdict = result

    print(globalrdf)
    print(rdfdict)
    print(countdict)

    for i in range(len(globalrdf)):
        tmp = 0
        for key, value in rdfdict[0].items():
            tmp += value[i]
        ReportTest(("Sum of partial RDFs matches global one (%d)" % (i,)),
                tmp, globalrdf[i], 0)


testtypes_NaCl = ((5.1, 6.001, 100, False),
                  (5.2, 6.001, 100, True),
                  (5.1, 15.001, 100, False),
                  (5.15, 15.001, 100, True))

@serial
@pytest.mark.parametrize('latconst, maxrdf, nbins, withemt', testtypes_NaCl)
def test_raw_rdf_NaCl(latconst, maxrdf, nbins, withemt):

    atoms = NaCl(directions=[[1,0,0],[0,1,0],[0,0,1]], symbol=("Cu","Au"),
                 size=(10,10,10), latticeconstant=latconst, debug=0)
    natoms = len(atoms)
    ReportTest("Number of atoms", natoms, 8000, 0)

    if withemt:
        atoms.calc = EMT()
        print(atoms.get_potential_energy())

    result = _asap.RawRDF(atoms, maxrdf, nbins, zeros(len(atoms), int32), 1,
                          ListOfElements(atoms))
    z = atoms.get_atomic_numbers()[0]

    globalrdf, rdfdict, countdict = result

    print(globalrdf)
    localrdf = zeros(globalrdf.shape)
    for i in (29,79):
        for j in (29,79):
            x = rdfdict[0][(i,j)]
            print("LOCAL", i, j)
            print(x)
            localrdf += x


    ReportTest("Local and global RDF are identical",
               min( globalrdf == localrdf), 1, 0)
    ReportTest("Atoms are counted correctly",
               countdict[0][29] + countdict[0][79], natoms, 0)
    
    shellpop = [6, 12, 8, 6, 24, -1]
    shell = [sqrt(i+1.0)/2.0 for i in range(6)]
    print(shell)
    print(shellpop)
    n = 0

    dr = maxrdf/nbins

    for i in range(nbins):
        if (i+1)*dr >= shell[n] * latconst:
            if shellpop[n] == -1:
                print("Reached the end of the test data")
                break
            ReportTest(("Shell %d (%d)" % (n+1, i)), globalrdf[i],
                       natoms*shellpop[n], 0)
            n += 1
        else:
            ReportTest(("Between shells (%d)" % (i,)), globalrdf[i], 0, 0)





from asap3 import EMT, MakeParallelAtoms, BundleTrajectory, units
from asap3.md.verlet import VelocityVerlet
from ase.lattice.cubic import FaceCenteredCubic
from asap3.io.trajectory import *
import sys, os, time
from asap3.mpi import world
from asap3.test.pytest_markers import ReportTest, parallel
import numpy as np

import pytest

ismaster = world.rank == 0
isparallel = world.size != 1
delete = True
precision = 1e-8

def maketraj(atoms, t, nstep):
    e = [atoms.get_potential_energy()]
    print("Shape of force:", atoms.get_forces().shape)
    dyn = VelocityVerlet(atoms, 5*units.fs)
    for i in range(nstep):
        dyn.run(10)
        energy = atoms.get_potential_energy()
        e.append(energy)
        if ismaster:
            print("Energy: ", energy)
        if t is not None:
            t.write()
    return e

def checktraj(t, e, cpus=None):
    i = 0
    for energy in e:
        atoms = t.get_atoms(i, cpus)
        atoms.calc = EMT()
        ReportTest("Checking frame %d / cpus=%s" % (i, str(cpus)),
                   atoms.get_potential_energy(), energy, precision)
        i += 1

@parallel
def test_split_bundletraj(in_tmp_dir, cpulayout):
    'Test writing of BundleTrajectory in "split mode" in parallel simulations.'
    if ismaster:
        initial = FaceCenteredCubic(size=(10,10,10), symbol="Cu", pbc=(1,0,0))
    else:
        initial = None
    if isparallel:
        atoms = MakeParallelAtoms(initial, cpulayout)
    else:
        atoms = initial.copy()
        
    atoms.calc = EMT()
    atoms.set_momenta(np.zeros((len(atoms), 3)))
    print("Writing trajectory")
    traj = BundleTrajectory("traj1.bundle", "w", atoms,
                            split=True, singleprecision=False)
    traj.write()
    energies = maketraj(atoms, traj, 10)
    traj.close()

    if ismaster:
        print("Reading trajectory (serial)")
        traj = BundleTrajectory("traj1.bundle")
        checktraj(traj, energies)

    if isparallel:
        world.barrier()
        print("Reading trajectory (parallel)")
        traj = BundleTrajectory("traj1.bundle")
        checktraj(traj, energies, cpulayout)
        world.barrier()

    print("Repeating simulation")
    atoms = traj.get_atoms(5, cpulayout)
    atoms.calc = EMT()
    energies2 = maketraj(atoms, None, 5)
    if ismaster:
        for i in range(5):
            ReportTest("Rerun[%d]" % (i,), energies2[i], energies[i+5], precision)
    traj.close()
    world.barrier()

    print("Appending to trajectory")
    atoms = BundleTrajectory("traj1.bundle").get_atoms(-1, cpulayout)
    atoms.calc = EMT()
    traj = BundleTrajectory("traj1.bundle", "a", atoms)
    energies2 = maketraj(atoms, traj, 5)
    traj.close()
    world.barrier()

    if ismaster:
        print("Reading longer trajectory")
        traj = BundleTrajectory("traj1.bundle")
        checktraj(traj, energies + energies2[1:])

    world.barrier()
    if delete:
        if ismaster:
            print("Deleting trajectory")
        BundleTrajectory.delete_bundle("traj1.bundle")



import numpy as np
from asap3.test.npt_module import main, bulkmodulus, MDLogger
from asap3.md.nptberendsen import NPTBerendsen

from asap3.test.pytest_markers import serial, parallel

import pytest

@pytest.mark.slow
@serial
def test_npt_berendsen():
    def berdynmaker(atoms, T0, p0, dt, *, taut, taup, rng, logint):
        dyn = NPTBerendsen(atoms, timestep=dt, temperature_K=T0, pressure_au=p0, 
                        taut=taut, taup=taup, compressibility_au=1 / bulkmodulus)
        log = MDLogger(dyn, atoms, '-', peratom=True, stress=True)
        dyn.attach(log, interval=logint)
        return dyn

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    main(berdynmaker, rng, failfluct=True)

@pytest.mark.slow
@parallel
def test_par_npt_berendsen():
    def berdynmaker(atoms, T0, p0, dt, *, taut, taup, rng, logint):
        dyn = NPTBerendsen(atoms, timestep=dt, temperature_K=T0, pressure_au=p0, 
                        taut=taut, taup=taup, compressibility_au=1 / bulkmodulus)
        log = MDLogger(dyn, atoms, '-', peratom=True, stress=True)
        dyn.attach(log, interval=logint)
        return dyn

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    #rng = np.random.default_rng()
    main(berdynmaker, rng, failfluct=True,  parallel=True)



from asap3 import *
from ase.lattice.cubic import *
from asap3.test.pytest_markers import serial


import pytest

@serial
@pytest.mark.core
def test_exc_illegal_element():
    "Illegal elements are an error."
    atoms = FaceCenteredCubic(size=(10, 10, 10), symbol="Cu")
    z = atoms.get_atomic_numbers()
    z[7]=8
    atoms.set_atomic_numbers(z)

    with pytest.raises(AsapError):
        atoms.calc = EMT()
    
@serial
@pytest.mark.core
def test_exc_atoms_collide():
    "Atoms on top of each other."
    atoms = FaceCenteredCubic(size=(10, 10, 10), symbol="Cu")
    r = atoms.get_positions()
    r[10] = r[11]
    atoms.set_positions(r)
    atoms.calc = EMT()

    with pytest.raises(AsapError):
        e = atoms.get_potential_energy()


@serial
@pytest.mark.core
def test_exc_array_size():
    "Atoms with malformed array (size)."
    atoms = FaceCenteredCubic(size=(10, 10, 10), symbol="Cu")
    atoms.arrays['numbers'] = atoms.arrays['numbers'][:len(atoms)//2]

    with pytest.raises(AsapError):
        atoms.calc = EMT()
        e = atoms.get_potential_energy()


@serial
@pytest.mark.core
def test_exc_array_type():
    "Atoms with malformed array (type)."
    atoms = FaceCenteredCubic(size=(10, 10, 10), symbol="Cu")
    atoms.arrays['positions'] = atoms.arrays['positions'].astype(int)

    with pytest.raises(AsapError):
        atoms.calc = EMT()
        e = atoms.get_potential_energy()




"""Test FIRE minimization algorithm and RestrictedCNA."""

from asap3 import EMT, MakeParallelAtoms
from asap3.mpi import world
from asap3.analysis.localstructure import RestrictedCNA

from asap3.optimize.fire import FIRE
from ase.build import bulk
from asap3.test.pytest_markers import ReportTest
import numpy as np

import pytest


def test_FIRE():

    ismaster = world.rank == 0
    isparallel = world.size != 1
    cpulayout = [1, 1, world.size]

    if ismaster:
        atoms = bulk('Cu')
        atoms = atoms.repeat((10, 10, 100))
        dx = 0.3 * np.sin(42 * np.arange(3 * len(atoms)))
        dx.shape = (-1, 3)
        atoms.set_array("perturbation", dx)
    else:
        atoms = None
    if isparallel:
        atoms = MakeParallelAtoms(atoms, cpulayout)    

    print("Testing FIRE relaxation")
    atoms.calc = EMT()
    e0 = atoms.get_potential_energy()
    natoms = atoms.get_global_number_of_atoms()
    ReportTest("Number of atoms", natoms, 10000, 0)
    print("Potential energy before perturbation:", e0, e0/natoms)

    atoms.set_positions(atoms.get_array("perturbation") + atoms.get_positions())
    e1 = atoms.get_potential_energy()
    print("Potential energy after perturbation:", e1, e1/natoms)

    dyn = FIRE(atoms)
    dyn.run(0.005)
    e2 = atoms.get_potential_energy()
    print("Potential energy after perturbation:", e2, e2/natoms)
    ReportTest("Energy after relaxation", e2, e0, 0.02)
    assert dyn.nsteps < 260, "Number of FIRE steps"

    RestrictedCNA(atoms)
    t = atoms.get_tags()
    nwrong = np.not_equal(t, 0).sum()
    ReportTest("Number of non-FCC atoms", nwrong, 0, 0)



"""Test for memory leaks in the EMT code."""

import asap3
from ase.lattice.compounds import L1_2
import numpy as np

steps = 50
msteps = 5

def makeatoms(causetrouble):
    atoms = L1_2(size=(10,10,10), symbol=('Au', 'Cu'), latticeconstant=4.0)
    r = atoms.get_positions()
    r += np.random.normal(0.0, 0.0001, r.shape)
    if causetrouble:
        r[100] = r[110]
    atoms.set_positions(r)
    return atoms

m0 = asap3.heap_mallinfo()
if m0 < 0:
    print("Memory monitoring not supported, test skipped.")
else:
    print("Memory at startup", m0, 'kB')
    
    for throwexception in (False, True):
        for replacepotential in (False, True):
            for replaceatoms in (False, True):
                print ("Running test: ReplacePotential=%s ReplaceAtoms=%s Exception=%s"
                       % (replacepotential, replaceatoms, throwexception))
                leak = 0
                if not replaceatoms:
                    atoms = makeatoms(throwexception)
                if not replacepotential:
                    pot = asap3.EMT()
                for i in range(steps):
                    if replaceatoms:
                        atoms = makeatoms(throwexception)
                    if replacepotential:
                        pot = asap3.EMT()
                    if i == 0 or replaceatoms or replacepotential:
                        atoms.calc = pot
                    try:
                        e = atoms.get_potential_energy()
                        #f = atoms.get_forces()
                    except asap3.AsapError:
                        pass
                    m = int(asap3.heap_mallinfo())
                    if i == msteps:
                        m0 = m
                    if i % msteps == 0 and i > msteps:
                        print("    Memory usage:", m, "kB")
                        if m > m0:
                            leak = max(leak, m - m0)
                if leak > 1:
                    print("      *** MEMORY LEAK DETECTED ***")

                            
                            



from asap3.Tools.ParameterOptimization.Optimization import ParameterOptimization
from asap3.Tools.ParameterOptimization.Search import ParameterSearch
from asap3.Tools.ParameterOptimization.Sampling import ParameterSampling
from asap3.Tools.ParameterOptimization.Performance import ParameterPerformance




# Copyright (C) 2025 Jakob Schiotz and Computational
# Atomic-scale Materials Design (CAMD), Department of Physics,
# Technical University of Denmark.  Email: schiotz@fysik.dtu.dk
#
# This file is part of Asap version 3.
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# version 3 as published by the Free Software Foundation.  Permission
# to use other versions of the GNU Lesser General Public License may
# granted by Jakob Schiotz or the head of department of the
# Department of Physics, Technical University of Denmark, as
# described in section 14 of the GNU General Public License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# and the GNU Lesser Public License along with this program.  If not,
# see <http://www.gnu.org/licenses/>.

import sys
import os
import string
import subprocess
import socket   # Just for the hostname
import argparse
from asap3 import __version__

asap_sbatch_doc = \
"""asap-sbatch: Submit an Asap job to Niflheim.

Usage:
    asap-sbatch [options] job.py [job-command-line]

where options are any options that could otherwise be passed to sbatch
and job-command-line is passed on to the job.  The job file MUST end in .py

In addition to the usual sbatch options, an option of the form --ASAP=X
may be passed, where X is one of the letters S, P or T specifying a
serial, parallel or multithreaded application (overriding the usual
detection).
"""

def asap_sbatch(prog='asap-sbatch', desc=asap_sbatch_doc, 
                version=__version__, args=None):
    "The asap-sbatch command"
    # Parsing of arguments is done manually, as we need to handle
    # a command line of the form
    #   asap-sbatch [arguments] script.py [script-arguments]
    # where the arguments may both be arguments to asap-sbatch
    # and unknown arguments passed on to the sbatch command.
    # argparse can help with this, but requires almost as much
    # manual work, so we just keep the orginal code from
    # scripts/asap-sbatch

    options = []
    job = []
    script = None
    asapargs = ""
    quiet = False
    private_production = False

    parse_partition = PartitionSanity()
                        
    # Parse the asap-sbatch command line
    if args == None:
        args = sys.argv[1:]
    for arg in args:
        if script is None:
            if arg.lower().endswith('.py'):
                script = arg
            else:
                if arg.upper().startswith("--ASAP="):
                    asapargs = arg[7:].upper()
                elif arg.lower() == "--quiet":
                    quiet = True
                elif arg.lower() == "--production":
                    # Developers only!
                    private_production = True
                else:
                    options.append(arg)
        else:
            job.append(arg)

    if script is None:
        raise ValueError("Cannot recognize the job script on the command line.  It must end with .py")

    # Construct the sbatch command line.
    sbatch = ["sbatch"]
    sbatch.extend(options)

    # Construct the job command line.  All options are single-quoted in
    # case they contain spaces or similar.
    jobcommand = script
    for arg in job:
        jobcommand = jobcommand + " '" + arg + "'"

    # Find default name for job
    defname = os.path.splitext(os.path.basename(script))[0]
    assert script.endswith(defname+'.py')
    # Remove weird characters
    for i, c in enumerate(defname):
        if not c in string.ascii_letters+string.digits:
            defname = defname[:i] + "_" + defname[i+1:]

    # Parse the script, collect any #SBATCH lines.
    slurmlines = ["#SBATCH -J "+defname+"\n"]
    for line in open(script):
        if line.startswith("#SBATCH"):
            slurmlines.append(line)
            if '--partition' in line:
                parse_partition(line)
        elif line.startswith("#PBS"):
            print("WARNING: PBS job control line found: ", line.strip())
            print("Passing it on to sbatch and hoping for the best ...")
            slurmlines.append(line)

    # Look for '--partition' arguments on the command line
    for opt in sbatch:
        if '--partition' in opt:
            parse_partition(opt)

    # Check that the last found partition specification is compatible with the
    # submitting node
    parse_partition.check()

    # Construct the script to be submitted.
    submitscript = "#!/bin/bash -l\n"
    for line in slurmlines:
        submitscript += line

    venv = os.getenv('VIRTUAL_ENV')
    if venv:
        if not quiet:
            print("Virtual environment detected:", venv)
        submitscript += 'source "{:s}/bin/activate"'.format(venv)
        
    if private_production:
        # Developers only!
        if not quiet:
            print("Production mode enabled on Niflheim7")
        submitscript += "export PYTHONPATH=`echo $PYTHONPATH | sed 's/development/production/g'`\n"
        submitscript += "export PATH=`echo $PATH | sed 's/development/production/g'`\n"

    submitscript += "\n"
    #submitscript += "mpirun -mca pml cm -mca mtl psm2 asap-python {0}\n".format(jobcommand)
    submitscript += 'echo "*** Simulation started at: `date`"\n' 
    submitscript += "mpirun python3 {0}\n".format(jobcommand)
    submitscript += 'echo "*** Simulation ended at: `date`"\n' 

    if "ASAPSBATCHVERBOSE" in os.environ and not quiet:
        print("Submitting job:")
        for line in submitscript.split("\n"):
            print("   ", line)
        print("Submitting with the command: ", " ".join(sbatch))
    print()

    sbatchproc = subprocess.Popen(sbatch, stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE, close_fds=True)
    (out, err) = sbatchproc.communicate(submitscript.encode())
    errcode = sbatchproc.wait()
    if errcode:
        print("sbatch failed with error code", str(errcode), file=sys.stderr)
        print("Command line:", sbatch, file=sys.stderr)
        print("Standard error of command:")
        print(err.decode(errors='replace'))
        sys.exit("sbatch failed")
    print(out.decode())    

# Partition sanity check
class PartitionSanity:
    """Check that the submitting host and the SLURM partition match.

    On some SLURM installations, the architecture of the submitting node
    must match the architecture of the desired computate node(s).
    
    If the hostname of the submitting node is in the table below, 
    the partition name must be correct.  Unknown submit nodes always
    pass the test.
    """
    #expected_partitions = {'sylg.fysik.dtu.dk': ['xeon24', 'xeon24_512',
    #                                             'xeon24_test'],
    #                       'thul.fysik.dtu.dk': ['xeon16', 'xeon16_128', 
    #                                             'xeon16_256'],
    #                       'fjorm.fysik.dtu.dk': ['xeon8']}
    expected_partitions = {}
    def __init__(self):
        self.part = "<default>"
    
    def __call__(self, s):
        words = s.split()
        for w in words:
            if w.startswith('--partition='):
                _, part = w.split('=')[:2]
                self.part = part
    
    def check(self):
        hostname = socket.getfqdn()
        if hostname in self.expected_partitions:
            expected = self.expected_partitions[hostname]
            if self.part not in expected:
                print("You are submitting from the wrong login node:", file=sys.stderr)
                print("   Login node:", hostname, file=sys.stderr)
                print("   Matching partitions:", str(expected), file=sys.stderr)
                print("   Specified partition:", self.part, file=sys.stderr)
                print("", file=sys.stderr)
                print("Please submit this job from login node", file=sys.stderr)
                found = False
                for k, v in self.expected_partitions.items():
                    if self.part in v:
                        print("  ", k, file=sys.stderr)
                        found = True
                if not found:
                    print("   <no acceptable login node found>")
                    print()
                    print("Does the partition '{}' exist?".format(self.part))
                sys.exit(1)





if 0:
    parser = argparse.ArgumentParser(prog=prog, description=desc)
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s-{version}')
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="Don't print stuff, just do it.")
    parser.add_argument('--ASAP', action="extend", nargs="+", type=str)
    # parser.add_argument('script')
    # parser.add_argument("script_args", nargs=argparse.REMAINDER)

    args, unknown = parser.parse_known_args(args)
    # Now unknown is a list of unparsed arguments, including the script name.
    # Any argument bef

    script = args.script
    if not script.endswith('.py'):
        raise ValueError('Cannot recognize the job script on the command line.'
                         '  It must end with .py')

    # Construct the job command line.  All options are single-quoted in
    # case they contain spaces or similar.
    sbatch = ["sbatch"]
    jobcommand = script
    for arg in args.script_args:
        jobcommand = jobcommand + " '" + arg + "'"

    # Find default name for job
    defname = os.path.splitext(os.path.basename(script))[0]
    assert script.endswith(defname+'.py')
    # Remove weird characters
    for i, c in enumerate(defname):
        if not c in string.ascii_letters+string.digits:
            defname = defname[:i] + "_" + defname[i+1:]




# VelocityDistributions.py -- set up a velocity distribution

"""Module for setting up e.g. Maxwell-Boltzmann velocity distributions.

Currently, only one function is defined, MaxwellBoltzmannDistribution,
which sets the momenta of a list of atoms according to a
Maxwell-Boltzmann distribution at a given temperature.
"""

import numpy as np
import asap3.mpi
from ase.md.velocitydistribution import _maxwellboltzmanndistribution
from ase.md.velocitydistribution import ZeroRotation as _ZeroRotation
from ase.md.md import process_temperature
import ase

def MaxwellBoltzmannDistribution(atoms, temp=None, *, temperature_K=None,
                                     force_temp=False, rng=None):
    """Sets the momenta to a Maxwell-Boltzmann distribution.

    Parameters:

    atoms: Atoms object
        The atoms.  Their momenta will be modified.

    temp: float (deprecated)
        The temperature in eV.  Deprecated, used temperature_K instead.

    temperature_K: float
        The temperature in Kelvin.

    force_temp: bool (optinal, default: False)
        If True, random the momenta are rescaled so the kinetic energy is 
        exactly 3/2 N k T.  This is a slight deviation from the correct
        Maxwell-Boltzmann distribution.

    rng: Numpy RNG (optional)
        Random number generator.  Default: numpy.random
    """
    temp = ase.units.kB * process_temperature(temp, temperature_K, 'eV')

    try:
         momenta = _maxwellboltzmanndistribution(atoms.get_masses(),
                                                temp=temp,
                                                comm=ase.parallel.DummyMPI(),
                                                rng=rng)
    except TypeError:
        momenta = _maxwellboltzmanndistribution(atoms.get_masses(),
                                                temp=temp,
                                                communicator='serial',
                                                rng=rng)
    atoms.set_momenta(momenta)
    if force_temp:
        force_temperature(atoms, temp, 'eV')

def Stationary(atoms, preserve_temperature=True):
    "Sets the center-of-mass momentum to zero."

    # Save initial temperature
    temp0 = atoms.get_temperature()

    p = atoms.get_momenta()
    p0 = np.sum(p, 0)
    # We should add a constant velocity, not momentum, to the atoms
    m = atoms.get_masses()
    mtot = np.sum(m)
    if getattr(atoms, "parallel", False):
        data = np.zeros(4, float)
        data[:3] = p0
        data[3] = mtot
        asap3.mpi.world.sum(data)
        p0 = data[:3]
        mtot = data[3]
    v0 = p0/mtot
    p -= v0*m[:,np.newaxis]
    atoms.set_momenta(p)

    if preserve_temperature:
        force_temperature(atoms, temp0)

# We cannot just use the ASE version of ZeroTemperature, as it is calling force_temperature,
# which would then be the ASE version and not the Asap version.
def ZeroRotation(atoms, preserve_temperature=True):
    "Sets the total angular momentum to zero by counteracting rigid rotations."

    # Save initial temperature
    temp0 = atoms.get_temperature()
    _ZeroRotation(atoms, preserve_temperature=False)
    if preserve_temperature:
        force_temperature(atoms, temp0)

def force_temperature(atoms, temperature, unit="K"):
    """ force (nucl.) temperature to have a precise value

    Parameters:
    atoms: ase.Atoms
        the structure
    temperature: float
        nuclear temperature to set
    unit: str
        'K' or 'eV' as unit for the temperature
    """

    eps_temp = 1e-12  # define a ``zero'' temperature to avoid divisions by zero

    if unit == "K":
        E_temp = temperature * ase.units.kB
    elif unit == "eV":
        E_temp = temperature
    else:
        raise ValueError(f"Unit '{unit}' is not supported, use 'K' or 'eV'.")

    if temperature > eps_temp:
        E_kin0 = atoms.get_kinetic_energy() / atoms.get_global_number_of_atoms() / 1.5
        gamma = E_temp / E_kin0
    else:
        gamma = 0.0
    atoms.set_momenta(atoms.get_momenta() * np.sqrt(gamma))




"""multidislocation - a module for setting up multiple dislocations."""


import itertools
import numpy as np

def fcc_slipsystems(withlines=False):
    """Return the different slip systems in FCC.

    Returns a list with the 12 slip systems.  Each system is
    a slip plane and a double Burgers vector.  The double of
    Burgers vector is used since that consists of integers.

    If the optional argument withlines is True, then each
    combination of a slip system and one of 6 possible
    dislocation line vectors are chosen, giving 72 "extended
    slip systems" instead.
    """
    # Four kinds of {111} planes
    planes111 = [(1,1,1), (1,1,-1), (1,-1,1), (1,-1,-1)]
    # Find all possible <110> directions
    temp = (set(itertools.permutations((1,1,0)))
            | set(itertools.permutations((1,-1,0)))
            | set(itertools.permutations((-1,-1,0)))
            )
    # But keep only the ones where the first nonzero index is positive
    direct110 = [x for x in temp if x > (0,0,0)]
    direct110.sort()
    direct110_signed = [x for x in temp]
    direct110_signed.sort()
    
    slipsystems = []
    for plane in planes111:
        for direction in direct110:
            if np.dot(plane, direction) == 0:
                if not withlines:
                    # Just use the slip system
                    slipsystems.append((plane, direction))
                else:
                    # Loop over possible dislocation lines as well
                    for dislline in direct110_signed:
                        if np.dot(plane, dislline) == 0:
                            slipsystems.append((plane, direction, dislline))
    if withlines:
        assert len(slipsystems) == 72
    else:
        assert len(slipsystems) == 12
    return slipsystems

def random_in_sphere(radius):
    """Returns a random point inside sphere of a given radius."""
    while True:
        candidate = np.random.uniform(-1,1,3)
        if np.dot(candidate, candidate) < 1.0:
            return radius * candidate

def random_dislocations(radius, density, verbose=True):
    """Create a random collection of dislocations.

    The dislocations all pass through a sphere with a
    given radius (the first parameters, in Angstrom).

    The dislocation density inside the sphere is given
    by the second parameter (in SI units, i.e. m^-2).

    This function returns parameters for creating the dislocations:
    a list of (point on line, Burgers vector, direction).
    """
    # Convert the density to A^-2:
    density /= 1e20
    cross_section = np.pi * radius**2
    avg_num = density * cross_section
    actual_num = np.random.poisson(avg_num)
    if verbose:
        print("Average number of dislocations:", avg_num)
        print("Actual number of dislocations:", actual_num)
    slipsys = fcc_slipsystems(True)
    result = []
    for i in range(actual_num):
        point = random_in_sphere(radius)
        plane, burgers, line = slipsys[np.random.randint(len(slipsys))]
        result.append((point, 0.5*np.array(burgers), line))
    return result


if __name__ == "__main__":
    from ase.lattice.cubic import FaceCenteredCubic
    from asap3.setup.dislocation import Dislocation
    from asap3 import EMT
    from asap3.analysis import CNA
    from ase.io import write
    import ase.data
    
    metal = 29  # Copper
    a0 = ase.data.reference_states[metal]['a']
    system = FaceCenteredCubic(metal, size=(30,30,30), latticeconstant=a0)
    system.set_pbc(False)
    size = system.get_cell()[0,0]
    print(f"System size: {size} A;  Number of atoms: {len(system)}")
    center = 0.5 * size * np.ones(3)
    disloc = random_dislocations(size * np.sqrt(3), 3e16)
    field = None
    for point, burgers, line in disloc:
        print(point, burgers, line)
        new = Dislocation(point+center, line, a0 * burgers)
        if field is None:
            field = new
        else:
            field = field + new
    # Apply the field
    field.apply_to(system)
    write("disl.xyz", system)
    system.set_calculator(EMT())
    t = CNA(system)
    system.set_tags(t)
    write("disl.traj", system)
    sys2 = system[t != 0]
    write("disl2.traj", sys2)
    
    
        
        



'''Check safe migration triggered by secondary neighbor list.

Check that doing an analysis in a situation where this would
trigger a migration does not cause subsequent calculations to
fail (issue #43).
'''

from asap3 import *
from ase.lattice.cubic import FaceCenteredCubic
from asap3.md.velocitydistribution import *
from asap3.analysis import CoordinationNumbers, PTM
from ase.parallel import world
from ase.visualize import view
from asap3.test.pytest_markers import ReportTest, parallel

import pytest

# Important: always at least two cores along x direction
ismaster = world.rank == 0
isparallel = world.size != 1


def make_atoms(layout):
    """Make an atomic system.

    Some vacuum is added, so that shearing it changes the number of
    atoms on the different MPI tasks.
    """
    if ismaster:
        atoms = FaceCenteredCubic('Cu', size=(20,20,20))
        atoms.pbc = False
        atoms.center(vacuum=5.0, axis=0)
        atoms.pbc = True
    else:
        atoms = None
    if world.size > 1:
        atoms = MakeParallelAtoms(atoms, layout)
    atoms.calc = EMT()
    atoms.get_forces()
    return atoms

def deform_atoms(atoms):
    """Shear the atoms."""
    r = atoms.get_positions()
    df =  0.1 * r[:,1]
    r[:,0] += df
    atoms.set_positions(r)
    
@parallel
@pytest.mark.core
def test_safe_nbl(cpulayout):
    atoms = make_atoms(cpulayout)
    e0 = atoms.get_potential_energy()
    deform_atoms(atoms)
    e1 = atoms.get_potential_energy()
    if world.rank == 0:
        print(f"\n\nEnergy of atoms changed from {e0} to {e1}")
    assert np.abs(e0 - e1) > 50.0, 'Energy has changed significantly'

    atoms = make_atoms(cpulayout)
    print(f"A: {len(atoms)} atoms on cpu {world.rank}")
    deform_atoms(atoms)
    #atoms.get_potential_energy()    # Workaround
    CoordinationNumbers(atoms)
    atoms.get_forces()
    e2 = atoms.get_potential_energy()
    print(f"B: {len(atoms)} atoms on cpu {world.rank}")
    ReportTest("Energy after CoordinationNumbers()", e2, e1, 0.001)


    atoms = make_atoms(cpulayout)
    deform_atoms(atoms)
    #atoms.get_potential_energy()    # Workaround
    nbl = atoms.calc.get_neighborlist()
    nbl.check_and_update(atoms)
    atoms.get_forces()
    e2 = atoms.get_potential_energy()
    ReportTest("Energy after borrowed neighborlist update", e2, e1, 0.001)

    atoms = make_atoms(cpulayout)
    deform_atoms(atoms)
    #atoms.get_potential_energy()    # Workaround
    PTM(atoms, cutoff=4.7)
    atoms.get_forces()
    e2 = atoms.get_potential_energy()
    ReportTest("Energy after PTM()", e2, e1, 0.001)


    # The following are not expected to be affected by issue #43.

    for i, nbtype in enumerate((FullNeighborList, NeighborCellLocator)):
        atoms = make_atoms(cpulayout)
        nbl = nbtype(4.8, atoms)
        deform_atoms(atoms)
        #atoms.get_potential_energy()    # Workaround
        nbl.check_and_update(atoms)
        atoms.get_forces()
        e2 = atoms.get_potential_energy()
        ReportTest(f"Energy neighbor list type {i}", e2, e1, 0.001)



from ase.md.npt import NPT as aseNPT
from asap3.md.md import ParallelMolDynMixin
from asap3.mpi import world
from ase.utils import basestring
import sys

class NPT(aseNPT, ParallelMolDynMixin):
    def __init__(self, atoms, *args, **kwargs):
        ParallelMolDynMixin.__init__(self, 'NPT', atoms)
        aseNPT.__init__(self, atoms, *args, **kwargs)

    def _getnatoms(self):
        """Get the number of atoms.

        In a parallel simulation, this is the total number of atoms on all
        processors.
        """
        return self.atoms.get_global_number_of_atoms()

    def get_center_of_mass_momentum(self):
        "Get the center of mass momentum."
        cm = self.atoms.get_momenta().sum(0)
        world.sum(cm)
        return cm

    @classmethod
    def read_from_trajectory(cls, trajectory,  frame=-1, cpus=None, extra=None):
        """Read dynamics and atoms from trajectory (Class method).

        Simultaneously reads the atoms and the dynamics from a BundleTrajectory,
        including the internal data of the NPT dynamics object (automatically
        saved when attaching a BundleTrajectory to an NPT object).

        Arguments::

        trajectory
            The filename or an open asap3.io.BundleTrajectory object.

        frame (optional)
            Which frame to read.  Default: the last.

        cpus (optional)
            If specified, the atoms are read in parallel and distributed
            for a parallel simulation.  This argument should then be the
            cpu layout (a tuple of three integers).
        """
        if isinstance(trajectory, basestring):
            if trajectory.endswith('/'):
                trajectory = trajectory[:-1]
            if trajectory.endswith('.bundle'):
                from asap3.io.bundletrajectory import BundleTrajectory
                trajectory = BundleTrajectory(trajectory)
            else:
                raise ValueError("Cannot open '%': unsupported file format" % trajectory)
        atoms = trajectory.get_atoms(frame, cpus, extra=extra)
        # The original read_from_trajectory method does the rest of the
        # work.  As it is a class method, the unbound method is not available
        # in the usual way, but as the im_func attribute.  Change to __func__
        # in Python 3.X
        return aseNPT.read_from_trajectory.__func__(cls, trajectory,
                                                   frame=frame, atoms=atoms)

    # Store these arrays on the atoms
    q = property(lambda s: s.get("q"), lambda s, x: s.set("q", x))
    q_past = property(lambda s: s.get("q_past"), lambda s, x: s.set("q_past", x))
    q_future = property(lambda s: s.get("q_future"), lambda s, x: s.set("q_future", x))



import numpy as np

from ase import Atoms, units
from ase.data import chemical_symbols, atomic_numbers
from ase.md import VelocityVerlet
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.lattice.cubic import SimpleCubic
from asap3 import Morse
from asap3.test.pytest_markers import ReportTest, serial

import pytest

#from calculators.morse import MorsePotential as Morse

# Define constants and calculator
@pytest.fixture
def elements():
    return np.array([atomic_numbers['Ru'], atomic_numbers['Ar']])

@pytest.fixture
def epsilon():
    return np.array([[5.720, 0.092], [0.092, 0.008]])

@pytest.fixture
def alpha():
    return np.array([[1.475, 2.719], [2.719, 1.472]])

@pytest.fixture
def rmin():
    return np.array([[2.110, 2.563], [2.563, 4.185]])

@pytest.fixture
def rcut(rmin, alpha):
    return rmin.max() + 6.0 / alpha.min()

@serial
@pytest.mark.core
def test_potential_cutoff(elements, epsilon, alpha, rmin, rcut):
    for e1 in elements:
        for e2 in elements:
            calc = Morse(elements, epsilon, alpha, rmin)
            atoms = Atoms([e1, e2], [[0.0, 0.0, 0.0], [rcut + 1.0, 0.0, 0.0]])
            atoms.center(vacuum=1.0)
            atoms.calc = calc

            energy = atoms.get_potential_energy()

            s1, s2 = chemical_symbols[e1], chemical_symbols[e2]
            ReportTest(f"Energy for {s1}-{s2} with r > rcut",
                       energy, 0.0, 1e-12)

@serial
@pytest.mark.core
def test_potential_minimum(elements, epsilon, alpha, rmin, rcut):
    for i, e1 in enumerate(elements):
        for j, e2 in enumerate(elements):
            calc = Morse(elements, epsilon, alpha, rmin)
            atoms = Atoms([e1, e2], [[0.0, 0.0, 0.0], [rmin[i, j], 0.0, 0.0]])
            atoms.center(vacuum=1.0)
            atoms.calc = calc

            energy = atoms.get_potential_energy()

            s1, s2 = chemical_symbols[e1], chemical_symbols[e2]
            ReportTest(f"Energy for {s1}-{s2} with r = rmin",
                       energy, -epsilon[i, j], 1e-2)

@serial
def test_energy_conservation(elements, epsilon, alpha, rmin, rcut):
    calc = Morse(elements, epsilon, alpha, rmin)
    atoms = SimpleCubic('Ar', size=(10,10,10), latticeconstant=5.0)
    n = 0
    while n < 100:
        i = np.random.randint(len(atoms)-1)
        if atoms[i].number != atomic_numbers['Ru']:
            atoms[i].number = atomic_numbers['Ru']
            n += 1
    atoms.calc = calc

    # Set initial momentum
    MaxwellBoltzmannDistribution(atoms, temperature_K=300)

    # Run dynamics
    dyn = VelocityVerlet(atoms, 1.0 * units.fs, logfile='test-energy.dat', loginterval=10)
    dyn.run(10)
    etot = (atoms.get_potential_energy() + atoms.get_kinetic_energy())/len(atoms)
    print("%-9s %-9s %-9s" % ("Epot", "Ekin", "Sum"))
    for i in range(25):
        if i:
            dyn.run(100)
        epot = atoms.get_potential_energy()/len(atoms)
        ekin = atoms.get_kinetic_energy()/len(atoms)
        print(f"{epot:9.5f} {ekin:9.5f} {epot+ekin:9.5f}")
        ReportTest("Step %i." % (i,), epot+ekin, etot, 1e-3)



import numpy as np
from asap3.test.nvt_module import main
from asap3.md.nvtberendsen import NVTBerendsen
import ase

from asap3.test.pytest_markers import serial, parallel
from ase.parallel import world

import pytest

@pytest.mark.slow
@serial
@pytest.mark.skipif(ase.__version__ < '3.25.0', reason='ASE is too old.')
def test_nvt_nosehoover():
    from asap3.md.nose_hoover_chain import NoseHooverChainNVT

    def nhdynmaker(atoms, T0, dt, tau, rng, logint):
        return NoseHooverChainNVT(atoms, dt, temperature_K=T0, tdamp=tau, logfile='-', loginterval=logint)

    def berdynmaker(atoms, T0, dt, tau, rng, logint):
        return NVTBerendsen(atoms, timestep=dt, temperature_K=T0, taut=tau, logfile='-', loginterval=logint)

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    #rng = np.random.default_rng()

    main(nhdynmaker, rng, initdyn=berdynmaker)

@pytest.mark.slow
@parallel
@pytest.mark.skipif(ase.__version__ < '3.25.0', reason='ASE is too old.')
def test_par_nvt_nosehoover():
    from asap3.md.nose_hoover_chain import NoseHooverChainNVT

    def nhdynmaker(atoms, T0, dt, tau, rng, logint):
        return NoseHooverChainNVT(atoms, dt, temperature_K=T0, tdamp=tau, logfile='-', loginterval=logint)

    def berdynmaker(atoms, T0, dt, tau, rng, logint):
        return NVTBerendsen(atoms, timestep=dt, temperature_K=T0, taut=tau, logfile='-', loginterval=logint)

    seed = np.random.SeedSequence(2718281828459045)
    seed = seed.spawn(world.size)[world.rank]
    rng = np.random.default_rng(seed)   # Use a fixed seed for reproducability

    main(nhdynmaker, rng, initdyn=berdynmaker, parallel=True, sloppytime=True, failfluct=True)
    # Note: Fluctuation test always fail in parallel, as the boundaries
    # are free in order to test the migration better.  But that changes the
    # heat capacity, making the test fail.



from asap3 import *
from ase.optimize import QuasiNewton
import numpy as np
from asap3.test.pytest_markers import ReportTest, serial

import pytest


# Parameters for Ne-Ar system
# The off-diagonal values are calculated in the test function,
# so we can test the wrong way is caught.
elements = [10, 18]

epsilon = [[0.0036471, 0],
           [0, 0.0123529]]

sigma = [[1.0334400, 0],
         [0, 1.8887100]]

cutoff = 7.55

def runtest(atoms, elements, epsilon, sigma, cutoff, correct=True):
    # Make a copy of the parameters
    epsilon = np.array(epsilon)
    sigma = np.array(sigma)
    if correct:
        # Fill in the cross term
        for p in (epsilon, sigma):
            p[1,0] = np.sqrt(p[0,0] * p[1,1])
    else:
        # Fill in the cross term
        for p in (epsilon, sigma):
            p[0,1] = np.sqrt(p[0,0] * p[1,1])
    print("Epsilon:")
    print(epsilon)
    print("Sigma:")
    print(sigma)
    print("Symbols:", atoms.get_chemical_symbols())

    atoms.calc = LennardJones(elements, epsilon, sigma, cutoff, False)
    dyn = QuasiNewton(atoms)
    dyn.run(fmax=1e-6)
    energy = atoms.get_potential_energy()
    pos = atoms.positions
    diff = pos[1] - pos[0]
    distance = np.linalg.norm(diff)
    print("Distance:", distance)
    print("Energy:", energy)
    return distance, energy

@serial
def test_Ne2():
    a = Atoms(symbols="Ne2", positions=[[0,0,0], [1.8, 0, 0]], pbc=False)
    a.center(vacuum=5.0)
    d, e = runtest(a, elements, epsilon, sigma, cutoff)
    ReportTest("Energy of Ne2:", e, -epsilon[0][0], 1e-5)
    ReportTest("Distance of Ne2:", d, 2**(1/6) * sigma[0][0], 1e-4)

@serial
def test_Ar2():
    a = Atoms(symbols="Ar2", positions=[[0,0,0], [1.8, 0, 0]], pbc=False)
    a.center(vacuum=5.0)
    d, e = runtest(a, elements, epsilon, sigma, cutoff)
    ReportTest("Energy of Ar2:", e, -epsilon[1][1], 1e-5)
    ReportTest("Distance of Ar2:", d, 2**(1/6) * sigma[1][1], 1e-5)

@serial
def test_NeAr():
    a = Atoms(symbols="NeAr", positions=[[0,0,0], [1.8, 0, 0]], pbc=False)
    a.center(vacuum=5.0)
    d, e = runtest(a, elements, epsilon, sigma, cutoff)
    ReportTest("Energy of NeAr:", e, -np.sqrt(epsilon[0][0]*epsilon[1][1]), 1e-5)
    ReportTest("Distance of NeAr:", d, 2**(1/6) * np.sqrt(sigma[0][0]*sigma[1][1]), 1e-5)

@serial
def test_NeAr_wrong():
    'LennardJones must throw an exception on incorrect matrix.'
    a = Atoms(symbols="NeAr", positions=[[0,0,0], [1.8, 0, 0]], pbc=False)
    a.center(vacuum=5.0)
    with pytest.raises(ValueError):
        d, e = runtest(a, elements, epsilon, sigma, cutoff, correct=False)




import datetime
from ase.parallel import paropen

class Log:
	def __init__(self,name = "simulation_log"):
		d = datetime.date.today()
		t = datetime.datetime.now().time()
		
		self.date = str(d.day)+":"+str(d.month)+":"+str(d.year)
		self.time = str(t.hour)+":"+str(t.minute)
		
		self.filename = name+"_"+str(self.date)+"_"+str(self.time)+".txt"
		f = paropen(self.filename,"a")
		f.write("Log file containing parameters\n \n")
		f.close()
		
		
	def dumpstring(self,dumpstr):
		f = paropen(self.filename,"a")
		f.write(dumpstr+"\n")
		f.close()
		
	def dumpparamdict(self,paramdict):
		for p in paramdict:
			self.dumpstring(p+":"+str(paramdict[p]))
	
	
	def load2append(self,filetoload):
		f = open(filetoload)
		str_in = f.read()
		self.dumpstring(str_in)
		



import numpy as np

from asap3.Tools.ParameterOptimization.Optimization import ParameterOptimization
from asap3.Tools.ParameterOptimization.Search import ParameterSearch

class ParameterSampling(ParameterSearch):
    def __init__(self, elements, calc, initparam, varparam, stdparam, points,
                 quantities, latticeconstants, calc_args=None, debug=False):
        self.points = points
        ParameterSearch.__init__(self, elements, calc, initparam, varparam,
                                 stdparam, quantities, latticeconstants,
                                 calc_args, debug=0)

    def run(self, xtol=5e-4, ftol=2e0, delta=5e-2, dat=None, txt=None):
        self.nsteps = 0
        self.numberlist = []
        self.errorlist = []
        self.fitparlist = []
        self.optparlist = []
        self.optvallist = []

        fitparam_list = []
        self.sampling(fitparam_list)
        #print len(fitparam_list)
        #for pars in fitparam_list[163:175]:
        #    for (k, p) in pars.items():
        #        print k, p
        #    print "-----"
        #return

        for fitpar in fitparam_list:
            self.nsteps += 1

            # Run optimization
            opt = ParameterOptimization(self.elements, self.calc, fitpar,
                                        self.varparam, self.values,
                                        self.latticeconstants, self.calc_args,
                                        debug=self.debug)
            (error, optpar, optval) = opt.fit(xtol, ftol, delta, txt)

            self.save(error, fitpar, optpar, optval)

        self.write(dat, 'sampling')

    def sampling(self, output, tempparam=None, n=1):
        # Which parameter should be altered
        varcount = 0
        for (key, var) in self.varparam.items():
            for i, v in enumerate(var):
                if v:
                    varcount += 1
                    if varcount == n:
                        k = key
                        m = i
        #print "Varying %s at position %i (%i out of %i)" % (k, m, n, varcount)

        # What is the "standard diviation"
        if isinstance(self.stdparam[k], list):
            std = self.stdparam[k][m]
        else:
            std = self.stdparam[k]

        # Vary the parameter
        ps = self.initparam[k][m] * np.linspace(1.0 - std, 1.0 + std, self.points)
        for p in ps:
            newparam = {}
            for key in self.initparam.keys():
                if tempparam is None:
                    newparam[key] = self.initparam[key][:]
                else:
                    newparam[key] = tempparam[key][:]
            newparam[k][m] = p
            if n < varcount:
                self.sampling(output, newparam, n+1)
            else:
                output.append(newparam)





import os
import math
import pickle
import numpy as np
import time

try:
    import matplotlib.pyplot as plt
    import pylab as pl
    _graphics = True
except:
    _graphics = False

from datetime import datetime
from ase.visualize.primiplotter import PrimiPlotter, GifFile
from ase.utils import basestring
from asap3.nanoparticle_mc.cluster import Cluster
from asap3.nanoparticle_mc import data

class SurfaceMonteCarlo:
    def __init__(self, symbol=None, layers=None, size=None,
                 latticeconstant=None, symmetry=None, debug=0):

        self.debug = debug
        self.set_calculator()

        #Find the atomic number
        if symbol is not None:
            if isinstance(symbol, basestring):
                self.atomic_number = data.atomic_numbers[symbol]
            else:
                self.atomic_number = symbol
        else:
            raise Warning('You must specify a atomic symbol or number!')

        #Find the crystal structure
        if symmetry is not None:
            if symmetry.lower() in ['bcc', 'fcc', 'hcp']:
                self.symmetry = symmetry.lower()
            else:
                raise Warning('The %s symmetry does not exist!' % symmetry.lower())
        else:
            self.symmetry = data.reference_states[self.atomic_number]['symmetry'].lower()

        if self.debug:
            print('Crystal structure:', self.symmetry)

        #Find the lattice constant
        if latticeconstant is None:
            if self.symmetry == 'fcc':
                self.lattice_constant = data.reference_states[self.atomic_number]['a']
            else:
                raise Warning('Cannot find the lattice constant ' +
                               'for a %s structure!' % self.symmetry)
        else:
            self.lattice_constant = latticeconstant

        if self.debug:
            print('Lattice constant(s):', self.lattice_constant)

        #Set the initial layers
        if layers is None:
            raise Warning('You must specify a set of surfaces.')
        else:
            self.init_layers = list(layers)
            
        #Find the size that the clusters will be fitted to
        if size is None:
            self.fitsize = self.get_size_from_layers(layers)
        else:
            self.fitsize = size

    def run(self, steps=1000, temp=1000, txt=None, smc=None):
        #Open the output textfile
        if txt is not None:
            if not isinstance(txt, basestring):
                raise Warning('You must specify a valid filename.')
            f = open(txt, 'w', 1)

        #Initialize the dataobject
        if smc is not None:
            smc = SurfaceMonteCarloData(smc,
                                        self.atomic_number,
                                        self.symmetry,
                                        self.lattice_constant,
                                        self.fitsize)
            smc.set_dimensions(data.lattice[self.symmetry]['surface_count'])
            self.data = smc

        #Initialize the MC simulation
        oldset = self.fit_to_size(self.init_layers)
        oldsize = self.get_size_from_layers(oldset)
        oldenergy = self.get_energy_from_layers(oldset, True)
        diameter = self.get_diameter_from_layers(oldset)

        if smc is not None:
            smc.append((oldenergy, oldset, oldsize, diameter, None))
        
        #Print header to outputfile
        if isinstance(f, file):
            print('Monte Carlo Simulation on Cluster Surface Parameters \n', file=f)
            print('Date: %s' % datetime.now().strftime("%A, %d %b %Y %H:%M:%S"), file=f)
            print('Material: %s' % data.chemical_symbols[self.atomic_number], file=f)
            print('Lattice constant: %s' % self.lattice_constant, file=f)
            print('Fitting size: %s' % self.fitsize, file=f)
            print('Temperature: %i K' % temp, file=f)
            print('Thermal energy: %.3f eV\n' % (data.kB * temp), file=f)
            print(('Iteration  Size   Energy   Energy/Atom  ' +
                         'Surface no.  Change  Accept  Probability'), file=f)
            print(('%7i %7i %9.3f %10.4f' %  
                         (0, oldsize, oldenergy * oldsize, oldenergy)), file=f)

        #Start simulation
        acceptances = 0
        accepted = False
        for i in range(1, steps + 1):
            #Choose change to make and do it
            id = self.choice(range(len(oldset)))
            change = self.choice([-1,1])

            newset = oldset[:]
            newset[id] += change
            newsize = self.get_size_from_layers(newset)

            #If layers = 0 there cannot be removed one and the change has to be real
            while (newset[id] < 1 and change < 0) or newsize == oldsize:
                id = self.choice(range(len(oldset)))
                #change = self.choice([-1,1])

                newset = oldset[:]
                newset[id] += change
                newsize = self.get_size_from_layers(newset)

            #Remove excess layers there have been removed
            c = self.get_cluster_from_layers(newset)
            newset = list(c.get_layers())
            del c

            #Fit the size and calculate energy and size
            newset = self.fit_to_size(newset, id)
            newsize = self.get_size_from_layers(newset)
            newenergy = self.get_energy_from_layers(newset, True)

            #Print MC step info to outputfile
            if isinstance(f, file):
                print(('%7i %7i %9.3f %10.4f %9i %10i' %  
                             (i, newsize, newenergy * newsize, newenergy, id, change)), end=' ', file=f)

            #Accept or reject the change
            de = (newenergy - oldenergy) * (newsize + oldsize) / 2

            if de < 0:
                p = 1.0
                accepted = True
            else:
                p = math.exp(- de / (data.kB * temp))
                pc = self.random()
                if p > pc: accepted = True

            #If accepted do...
            if accepted:
                accepted = False
                acceptances += 1

                oldset = newset[:]
                oldsize = newsize
                oldenergy = newenergy
                diameter = self.get_diameter_from_layers(oldset)

                if smc is not None:
                    smc.append((oldenergy, oldset, oldsize, diameter, None))

                if isinstance(f, file):
                    print('%8s %10.2f ' % ('Yes', p) + time.ctime(), file=f)
            else:
                if isinstance(f, file):
                    print('%8s %10.2f ' % ('No', p) + time.ctime(), file=f)

        if isinstance(f, file):
            print('\nDate: %s' % datetime.now().strftime("%A, %d %b %Y %H:%M:%S"), file=f)
            print('Acceptance ratio: %.2f' % (acceptances * 1.0 / steps), file=f)

        if smc is not None:
            smc.write()

    def fit_to_size(self, surfaces, restriction=None):
        oldset = list(surfaces)
        oldsize = self.get_size_from_layers(oldset)

        if self.debug > 1:
            print(42 * '-')
            print('Fitting surfaces to match the size %i.' % self.fitsize)
            print(42 * '-')
            print('Start: ' + str(oldset) + '\n')
            print('Iter  Size  Surface no.  Change')
            print('  0 %6i  Time is now: ' % oldsize + time.ctime())

        i = 0
        while True:
            i += 1

            #Chose a change to make in the right direction
            id = self.choice(range(len(oldset)))

            if oldsize < self.fitsize:
                change = 1
            elif oldsize > self.fitsize:
                change = -1
            else:
                newsize = None
                break

            newset = oldset[:]
            newset[id] += change
            newsize = self.get_size_from_layers(newset)

            #If layers = 0 there cannot be removed one, the change has to be real and
            #the restricted parameter is keeped fixed
            while (newset[id] < 1 and change < 0) or newsize == oldsize or id == restriction:
                id = self.choice(range(len(newset)))

                newset = oldset[:]
                newset[id] += change
                newsize = self.get_size_from_layers(newset)

            #Remove excess layers there have been removed
            c = self.get_cluster_from_layers(newset)
            newset = list(c.get_layers())
            del c

            if self.debug > 1:
                print('%3i %6i %7i %10i Time: ' % (i, newsize, id, change) + time.ctime())

            #If the wanted size is passed or reached then break
            if oldsize < self.fitsize and newsize > self.fitsize:
                break
            elif oldsize > self.fitsize and newsize < self.fitsize:
                break
            elif newsize == self.fitsize:
                break

            #Save the new paremeter set and start again
            oldset = newset[:]
            oldsize = newsize

        #Chosing between the last two configurations the one with the lowest energy
        if newsize is not None:
            oldenergy = self.get_energy_from_layers(oldset, True)
            newenergy = self.get_energy_from_layers(newset, True)

            if self.debug > 1:
                print('\nIter   Energy   Energy/Atom')
                print('%3i %9.2f %10.3f' % (i - 1, oldenergy * oldsize, oldenergy))
                print('%3i %9.2f %10.3f' % (i, newenergy * newsize, newenergy))
                print('Time is now: ' + time.ctime())

            if newenergy < oldenergy:
                chosenset = newset
            else:
                chosenset = oldset
        else:
            chosenset = oldset

        if self.symmetry == 'fcc':
            chosenset = list(data.fcc.surface_centering(chosenset))

        #Check for different flaws in the chosen set of layers
        if self.debug > 1:
            print('\nFinal: ' + str(chosenset) + '\n')
            print('Time is now: ' + time.ctime())

            if newsize is not None and newenergy < oldenergy:
                oldcluster = self.get_cluster_from_layers(newset)
            else:
                oldcluster = self.get_cluster_from_layers(oldset)

            if (np.array(chosenset) < 0).any():
                raise Warning('Negative number of layers at a surface!')

            newcluster = self.get_cluster_from_layers(chosenset)

            if len(oldcluster) != len(newcluster):
                raise Warning('Size does not match: %i != %i' % 
                               (len(oldcluster), len(newcluster)))

            if (np.array(chosenset) != newcluster.get_layers()).any():
                raise Warning('None existing layers in chosen set')

            if self.debug > 2:
                oldcluster.make_neighborlist()
                newcluster.make_neighborlist()

                oldtypes = np.bincount(oldcluster.get_types())
                newtypes = np.bincount(newcluster.get_types())

                if (oldtypes != newtypes).any():
                    raise Warning(f'Types does not match: {oldtypes} != {newtypes}')

        return chosenset

    def set_calculator(self, calc=None):
        self.calc = calc

    def get_calculator(self):
        return self.calc

    #Some useful functions
    def choice(self, data):
        return data[np.random.randint(len(data))]

    def random(self):
        return np.random.random()

    def get_cluster_from_layers(self, layers):
        return Cluster(symbol=self.atomic_number,
                       layers=layers,
                       symmetry=self.symmetry,
                       latticeconstant=self.lattice_constant)

    def get_size_from_layers(self, layers):
        c = self.get_cluster_from_layers(layers)
        return len(c)
        
    def get_energy_from_layers(self, layers, atoms=False):
        if self.calc is None:
            raise Warning('You must specify a calculator to calculate the energy.')

        c = self.get_cluster_from_layers(layers)
        c.set_calculator(self.calc)
        
        if atoms:
            return c.get_potential_energy() / len(c)
        else:
            return c.get_potential_energy()

    def get_diameter_from_layers(self, layers):
        """Makes an estimate of the cluster diameter based on the average
        distance between opposite layers"""
        surface_mapping = data.lattice[self.symmetry]['surface_mapping']
        surface_data = data.lattice[self.symmetry]['surface_data']
        surface_count = data.lattice[self.symmetry]['surface_count']

        d = 0.0
        for i, l1 in enumerate(layers):
            l2 = layers[surface_mapping[i]]
            dl = surface_data[i]['d'] * self.lattice_constant
            d += (l1 + l2) * dl / surface_count

        return d

class SurfaceMonteCarloData:
    def __init__(self, filename=None, symbol=None, symmetry=None,
                 latticeconstant=None, fitsize=None, debug=0):
        if isinstance(symbol, basestring):
            self.atomic_number = data.atomic_numbers[symbol]
        else:
            self.atomic_number = symbol

        self.debug = debug
        self.filename = filename
        self.symmetry = symmetry
        self.lattice_constant = latticeconstant
        self.fitsize = fitsize

        self.arrays = {}
        self.arrays['surfaces'] = np.empty(0, dtype=int)
        self.arrays['energy'] = np.empty(0, dtype=float)
        self.arrays['size'] = np.empty(0, dtype=int)
        self.arrays['diameter'] = np.empty(0, dtype=float)
        self.arrays['counts'] = np.empty(0, dtype=int)

    def set_dimensions(self, surface_count):
        self.arrays['surfaces'] = self.arrays['surfaces'].reshape((0, surface_count))

    def __len__(self):
        return len(self.arrays['energy'])

    def __getitem__(self, index=-1):
        energy = self.arrays['energy'][index]
        surfaces = self.arrays['surfaces'][index]
        size = self.arrays['size'][index]
        diameter = self.arrays['diameter'][index]
        counts = self.arrays['counts'][index]

        return (energy, surfaces, size, diameter, counts)

    def __delitem__(self, index=-1):
        mask = np.ones(len(self), bool)
        mask[index] = False

        for name, a in self.arrays.items():
            self.arrays[name] = a[mask]

    def __setitem__(self, index, values):
        if values[0] is not None:
            self.arrays['energy'][index] = values[0]

        if values[1] is not None:
            self.arrays['surfaces'][index] = np.array(values[1], dtype=int)

        if values[2] is not None:
            self.arrays['size'][index] = values[2]

        if values[3] is not None:
            self.arrays['diameter'][index] = values[3]

        if values[4] is not None:
            self.arrays['counts'][index] = values[4]

    def append(self, values):
        for name, a in self.arrays.items():
            if name == 'counts':
                b = np.ones((a.shape[0] + 1,) + a.shape[1:], a.dtype)
            else:
                b = np.zeros((a.shape[0] + 1,) + a.shape[1:], a.dtype)

            b[:a.shape[0]] = a
            self.arrays[name] = b

        self[-1] = values

    def extend(self, other):
        if not isinstance(other, SurfaceMonteCarloData):
            raise Warning('The added object is not a SurfaceMonteCarloData instance!')

        if len(self) == 0:
            self.set_dimensions(other.arrays['surfaces'].shape[1])

        for name, a in self.arrays.items():
            b = other.arrays[name]
            if a.shape[1:] != b.shape[1:]:
                raise Warning(f'Different shapes: {a.shape[1:]} != {b.shape[1:]}')
            c = np.zeros((a.shape[0] + b.shape[0],) + a.shape[1:], a.dtype)
            c[:a.shape[0]] = a
            c[a.shape[0]:] = b
            self.arrays[name] = c

    def write(self, filename=None):
        if filename is None:
            filename = self.filename

        if not isinstance(filename, basestring):
            raise Warning('You must specify a valid filename.')

        if os.path.isfile(filename):
            os.rename(filename, filename + '.bak')
        
        d = {'atomicnumber': self.atomic_number,
             'symmetry': self.symmetry,
             'latticeconstant': self.lattice_constant,
             'fitsize': self.fitsize}
             
        f = open(filename, 'wb')
        f.write('ParameterMonteCarlo')
        pickle.dump(d, f)
        pickle.dump(self.arrays, f)
        f.close()

    def read(self, filename=None):
        if filename is None:
            filename = self.filename

        if not isinstance(filename, basestring):
            raise Warning('You must specify a valid filename.')

        if not os.path.isfile(filename):
            raise Warning('The file specified do not exist.')
        
        f = open(filename, 'rb')

        try:
            if f.read(len('ParameterMonteCarlo')) != 'ParameterMonteCarlo':
                raise Warning('This is not a compatible file.')
            d = pickle.load(f)
            self.arrays = pickle.load(f)
        except EOFError:
            raise Warinig('Bad file.')
        
        f.close()

        if 'energies' in self.arrays:
            self.arrays['energy'] = self.arrays['energies']
            self.arrays['size'] = self.arrays['sizes']
            self.arrays['counts'] = np.zeros(len(self))
            del self.arrays['energies']
            del self.arrays['sizes']

        if 'multi' in self.arrays:
            self.arrays['counts'] = self.arrays['multi']
            del self.arrays['multi']

        self.atomic_number = d['atomicnumber']
        self.symmetry = d['symmetry']
        self.lattice_constant = d['latticeconstant']
        self.fitsize = d['fitsize']

    #Function to filter the data
    def filter_energy(self, excess):
        """Function that filters the data leaving only entries with
        energy lower than 'excess'."""

        e = self.get_energies()
        s = self.get_sizes()
        de = (e - e.min()) * s.mean()

        mask = de < excess
        for name, a in self.arrays.items():
            self.arrays[name] = a[mask]

    def filter_number(self, number):
        """Function that filters the data leaving only entries with
        the lowest energy in number less than 'number'."""

        if len(self) > number:
            e = self.get_energies()
            ind = e.argsort().argsort()
            mask = ind < number
            for name, a in self.arrays.items():
                self.arrays[name] = a[mask]

    def filter_surfaces(self):
        """Function that filters the data leaving only geometric
        different clusters."""

        symmetries = data.lattice[self.symmetry]['surface_symmetries']

        arrays_new = {}
        for name, a in self.arrays.items():
            arrays_new[name] = np.empty((0,) + a.shape[1:], dtype=a.dtype)
        lookup_table = {}
            
        for i, s in enumerate(self.arrays['surfaces']):
            f = -1

            for j, sym in enumerate(symmetries):
                if len(sym) != len(s):
                    raise Warning('Lenght of surfaces does not match with the symmetry.')

                s1 = s[sym]

                if self.symmetry == 'fcc':
                    s1 = data.fcc.surface_centering(s1)

                k = lookup_table.get(tuple(s1), None)

                if k is not None:
                    if f < 0 or f == k:
                        if self.debug:
                            print('Surfaces %i found as %i with symmetry %i' % (i, k, j))
                        f = k
                    else:
                        print('Surfaces #%i:\n%s' % (i, s))
                        print('Symmetry %i' % j)
                        print('Found as filtered surfaces %i and ealier as %i:' % (k, f))
                        print('{}\n{}'.format(list(s1), list(arrays_new['surfaces'][f])))
                        raise Warning('Surfaces with applied symmetry ' +
                                       'found for different filtered surfaces!')

            if f < 0:
                for name, a in arrays_new.items():
                    a1 = np.zeros((a.shape[0] + 1,) + a.shape[1:], a.dtype)
                    a1[:a.shape[0]] = a
                    a1[a.shape[0]:] = self.arrays[name][i]
                    arrays_new[name] = a1
                    length = len(a1)
                s_as_tuple = tuple(data.fcc.surface_centering(s))
                lookup_table[s_as_tuple] = length-1
            else:
                arrays_new['counts'][f] += 1

        self.arrays = arrays_new

    def get_multiplicity(self, i):
        symmetries = data.lattice[self.symmetry]['surface_symmetries']
        s = self.get_surfaces()[i]

        m = 0
        for i, sym in enumerate(symmetries):
            s1 = data.fcc.surface_centering(s[sym])
            if (s == s1).all(): m += 1

        return m

    def get_multiplicities(self):
        m = []
        for i in range(len(self)):
            m.append(self.get_multiplicity(i))

        return np.array(m, int)

    def get_surfaces(self):
        return self.arrays['surfaces'].copy()

    def get_energies(self):
        return self.arrays['energy'].copy()

    def get_sizes(self):
        return self.arrays['size'].copy()

    def get_diameters(self):
        return self.arrays['diameter'].copy()

    def get_counts(self):
        return self.arrays['counts'].copy()

    def get_cluster(self, i):
        return Cluster(symbol=self.atomic_number,
                       layers=self.get_surfaces()[i],
                       symmetry=self.symmetry,
                       latticeconstant=self.lattice_constant,
                       multiplicity=self.get_multiplicity(i))

    #Function to make a movie
    def make_movie(self, filename=None, frames=10, interval=None):
        if not isinstance(filename, basestring):
            raise Warning('The specified filename is not valid.')

        if isinstance(interval, list):
            if (interval[0] < 0 or interval[0] >= len(self) or
                interval[1] < 0 or interval[1] >= len(self)):
                raise Warning('Index out of range.')

            if interval[0] < interval[1]:
                start = interval[0]
                end = interval[1]
            else:
                start = interval[1]
                end = interval[0]
        elif interval is None:
            start = 0
            end = len(self) - 1
        else:
            raise Warning('Interval is not a list.')

        if frames < (end - start):
            delta = int(round((end - start) * 1.0 / frames))
        else:
            delta = 1

        images = range(start, end + 1, delta)

        if (end - images[-1]) > delta / 2.0:
            images.append(end - 1)
        else:
            images[-1] = end

        p = PrimiPlotter(self.get_cluster(images[0]))
        p.set_output(GifFile('SMCimg'))
        p.set_rotation([45,45,45])
        p.autoscale('now')
        p.autoscale('off')
        p.set_scale(0.8 * p.get_scale())
        p.update()
        #os.popen(('mogrify -draw \'text 10,10 "Frame %i of %i"\' SMCimg0000.gif' % 
        #          (images[0], len(self))))
        del images[0]
        
        for i, j in enumerate(images):
            p.update(self.get_cluster(j))
            #os.popen(('mogrify -draw \'text 10,10 "Frame %i of %i"\' SMCimg%04i.gif' %
            #          (j, len(self) - 1, i + 1)))

        os.popen('convert -delay 70 -loop 10 SMCimg*.gif ' + filename)
        os.popen('rm SMCimg*.gif -f')

    #OLD STUFF
    def filter_energy_old(self, number=None, energy=None):
        """Function that filters the data leaving only entries with
        energy lower the 'energy' and in a number that is lower or
        equal to 'number'."""

        if number is None:
            number = len(self)

        if energy is None:
            energy = self.arrays['energy'].max()

        arrays_new = {}
        for name, a in self.arrays.items():
            arrays_new[name] = np.empty((0,) + a.shape[1:], dtype=a.dtype)

        appended = 0
        for i, e in enumerate(self.arrays['energy']):
            if e <= energy and appended < number:
                appended += 1

                for name, a in arrays_new.items():
                    a1 = np.zeros((a.shape[0] + 1,) + a.shape[1:], a.dtype)
                    a1[:a.shape[0]] = a
                    a1[a.shape[0]:] = self.arrays[name][i]
                    arrays_new[name] = a1

            elif e <= energy and appended >= number:
                emax = arrays_new['energy'].max()
                imax = arrays_new['energy'].argmax()

                if e < emax:
                    for name, a in arrays_new.items():
                        a1 = np.zeros((a.shape[0] + 1,) + a.shape[1:], a.dtype)
                        a1[:a.shape[0]] = a
                        a1[a.shape[0]:] = self.arrays[name][i]
                        arrays_new[name] = a1

                    mask = np.ones(len(arrays_new['energy']), dtype=bool)
                    mask[imax] = False

                    for name, a in arrays_new.items():
                        arrays_new[name] = a[mask]

        self.arrays = arrays_new



from asap3 import EMT
try:
    from asap3 import MakeParallelAtoms
except ImportError:
     pass
from ase.lattice.cubic import FaceCenteredCubic
from asap3.test.pytest_markers import ReportTest
from ase.parallel import world
import numpy as np

import pytest

@pytest.mark.core
def test_calculation_required(cpulayout):
    if cpulayout:
        if world.rank == 0:
            size = tuple(5 * np.array(cpulayout))
            atoms = FaceCenteredCubic(size=size, symbol="Cu")
        else:
            atoms = None
        atoms = MakeParallelAtoms(atoms, cpulayout)
    else:
        atoms = FaceCenteredCubic(size=(5,5,5), symbol="Cu")
    pot = EMT()
    atoms.calc = pot

    for i in (1, 2):
        print("*** Pass", i)
        ReportTest("Energy required", pot.calculation_required(atoms, ["energy"]), 1, 0)
        ReportTest("Forces required", pot.calculation_required(atoms, ["forces"]), 1, 0)
        ReportTest("Stress required", pot.calculation_required(atoms, ["stress"]), 1, 0)
        ReportTest("Magmom required", pot.calculation_required(atoms, ["magmoms"]), 1, 0)
        e = atoms.get_potential_energy()
        ReportTest("Energy not required", pot.calculation_required(atoms, ["energy"]), 0, 0)
        ReportTest("Forces required (II)", pot.calculation_required(atoms, ["forces"]), 1, 0)
        f = atoms.get_forces()
        ReportTest("Energy not required (II)", pot.calculation_required(atoms, ["energy"]), 0, 0)
        ReportTest("Forces not required", pot.calculation_required(atoms, ["forces"]), 0, 0)
        ReportTest("Energy or forces not required",
                pot.calculation_required(atoms, ["energy", "forces"]), 0, 0)
        ReportTest("Energy or stress required",
                pot.calculation_required(atoms, ["energy", "stress"]), 1, 0)
        s = atoms.get_stress()
        ReportTest("Stress not required", pot.calculation_required(atoms, ["stress"]), 0, 0)

        r = atoms.get_positions()
        r[0,0] += 0.1
        atoms.set_positions(r)





"""Element data - 79 Au Gold"""

name = 'Gold'
symbol = 'Au'
symmetry = 'fcc'

energy_slope = -0.090
energy_intersection = -1.537

energy_types = {1: -2.618, #bulk
                2: -2.237, #100 surface
                3: -2.343, #110 surface (top) or 111-111 edge
                4: -2.369, #110 surface (bottom)
                5: -2.352, #111 surface
                6: -2.028, #100-110 edge
                7: -2.215, #100-111 edge
                }

data = {'name': name,
        'symbol': symbol,
        'symmetry': symmetry,
        'energy_slope': energy_slope,
        'energy_intersection': energy_intersection,
        'energy_types': energy_types,
        }




from asap3 import EMT, EMTStandardParameters
from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.compounds import B2
import numpy as np
from asap3.test.pytest_markers import ReportTest, serial

import pytest

def check(a1, e):
    a2 = a1.copy()
    a1.calc = EMT(EMTStandardParameters())
    a2.calc = EMT()
    ReportTest(f"Energy ({e})", a2.get_potential_energy(),
               a1.get_potential_energy(), 1e-8)
    f1 = a1.get_forces().ravel()
    f2 = a2.get_forces().ravel()
    ReportTest(f"Forces ({e})", max(np.fabs(f2-f1)), 0.0, 1e-8)

@serial
@pytest.mark.core
def test_standard():
    elements = ["Al", "Ni", "Cu", "Pd", "Ag", "Pt", "Au"]
    print("Comparing Default and Standard parameter providers.")
    for e in elements:
        atoms = FaceCenteredCubic(size=(4,4,4), symbol=e)
        check(atoms, e)
        
    for i in range(1, len(elements)):
        e1 = elements[i]
        e2 = elements[i-1]
        atoms = B2(size=(4,4,4), symbol=(e1,e2), latticeconstant=3.5, pbc=False)
        check(atoms, e1+e2)

        atoms = B2(size=(4,4,4), symbol=(e2,e1), latticeconstant=3.5, pbc=False)
        check(atoms, e2+e1)





from asap3 import LennardJones
from ase.lattice.compounds import B2
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest

@serial
def test_LJ_stress():
    'Test for an old bug where stresses were not always correct in LennardJones.'
    atoms = B2(size=(332,166,5), symbol=('Cu', 'Zr'), latticeconstant=2.7)
    lj = LennardJones((29,40), 0.2 * np.array([0.5, 1.0, 1.0, 0.5]),
                    np.array([1.236, 2.0, 2.0, 2.352]),
                    rCut=2.0 * 2.352)
    atoms.calc = lj

    # If the following line is included, the bug is not seen.
    #atoms.get_potential_energy()
    stress = atoms.get_stress()
    exp_tr = 9.30356046e-02
    for i in range(3):
        ReportTest("Diagonal elements of stress tensor", stress[i], exp_tr, 1e-6)
    for i in range(3,6):
        ReportTest("Off-diagonal elements of stress tensor", stress[i], 0.0, 1e-8)
    



from asap3 import MakeParallelAtoms
from ase.cluster.cubic import FaceCenteredCubic
from asap3.md.velocitydistribution import *
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution as ASE_MaxwellBoltzmannDistribution
from ase.parallel import world, DummyMPI
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest

@pytest.mark.core
def test_velocitydist(cpulayout):
    # We need an asymmetric nanoparticle
    surfaces = [(1, 0, 0), (1, 1, 1), (0, -1, 0), (-1, 0, 0)]
    layers = [9, 7, 7, 8]
    lc = 4.08000

    ismaster = world.rank == 0
    isparallel = world.size != 1

    ###  Test the MaxwellBoltzmann distribution in parallel
    if ismaster:
        atoms = FaceCenteredCubic('Au', surfaces, layers, latticeconstant=lc)
        atoms.center(vacuum=3.0)
    else:
        atoms = None

    if isparallel:
        atoms = MakeParallelAtoms(atoms, cpulayout)
    MaxwellBoltzmannDistribution(atoms, temperature_K=300, force_temp=True)
    ReportTest("Temperature of Maxwell-Boltzmann distribution", atoms.get_temperature(), 300, 1e-6)


    ### Test the following:
    ###    ParallelListOfAtoms.get_center_of_mass()
    ###    ParallelListOfAtoms.get_angular_momentum()
    ###    ParallelListOfAtoms.get_moments_of_inertial()
    ###    Stationary()
    ###    ZeroRotation()

    # Create an atoms object with a nonzero angular momentum and center of mass momentum

    if ismaster:
        atoms = FaceCenteredCubic('Au', surfaces, layers, latticeconstant=lc)
        atoms.center(vacuum=3.0)
        try:
            ASE_MaxwellBoltzmannDistribution(atoms, temperature_K=300, force_temp=True, comm=DummyMPI())
        except TypeError:
            ASE_MaxwellBoltzmannDistribution(atoms, temperature_K=300, force_temp=True, communicator='serial')
        ReportTest("Temperature of serial Maxwell-Boltzmann distribution", atoms.get_temperature(), 300, 1e-6)
        atoms[0].momentum += np.array((0, 1.0, 0))
        init_com = atoms.get_center_of_mass()
        init_am = atoms.get_angular_momentum()
        init_mi = atoms.get_moments_of_inertia()
    else:
        atoms = None

    if isparallel:
        atoms = MakeParallelAtoms(atoms, cpulayout)
    temp = atoms.get_temperature()
    com = atoms.get_center_of_mass()
    am = atoms.get_angular_momentum()
    mi = atoms.get_moments_of_inertia()
    if ismaster:
        for i, ax in enumerate('xyz'):
            ReportTest(f"Center of mass ({ax})", com[i], init_com[i], 1e-6)
            ReportTest(f"Angular momentum ({ax})", am[i], init_am[i], 1e-6)
            ReportTest(f"Moments of inertial ({ax})", mi[i], init_mi[i], 1e-6)
        print("Angular momentum:", am)
        
    Stationary(atoms)
    ReportTest("Temperature after Stationary(atoms)", atoms.get_temperature(), temp, 1e-6)
    ZeroRotation(atoms)
    ReportTest("Temperature after ZeroRotation(atoms)", atoms.get_temperature(), temp, 1e-6)
    am = atoms.get_angular_momentum()
    for i, ax in enumerate('xyz'):
        ReportTest(f"Angular momentum ({ax}) is zero", am[i], 0, 1e-6)

